import pytest
from ckan.tests import factories
from ckan.common import g
from ckanext.attribution.lib import helpers


@pytest.mark.filterwarnings('ignore::sqlalchemy.exc.SADeprecationWarning')
@pytest.mark.ckan_config('ckan.plugins', 'attribution')
@pytest.mark.usefixtures('clean_db', 'with_plugins', 'with_request_context')
def test_can_edit_non_superuser():
    user = factories.User()
    g.user = user['name']
    assert not helpers.can_edit()


@pytest.mark.filterwarnings('ignore::sqlalchemy.exc.SADeprecationWarning')
@pytest.mark.ckan_config('ckan.plugins', 'attribution')
@pytest.mark.usefixtures('clean_db', 'with_plugins', 'with_request_context')
def test_can_edit_superuser():
    user = factories.Sysadmin()
    g.user = user['name']
    assert helpers.can_edit()



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit
from ckanext.attribution.logic.actions.helpers import (
    parse_contributors,
    get_author_string,
)
from ckanext.attribution.model.crud import (
    AgentContributionActivityQuery,
    AgentQuery,
    ContributionActivityQuery,
    PackageContributionActivityQuery,
    AgentAffiliationQuery,
)
from ckantools.decorators import action, basic_action
from ckanext.attribution.logic.actions.meta import help, schema


@action(schema.agent_affiliation_create, help.agent_affiliation_create)
def agent_affiliation_create(context, original_data_dict, agent_a_id, agent_b_id):
    for agent_id in [agent_a_id, agent_b_id]:
        try:
            toolkit.get_action('agent_show')(context, {'id': agent_id})
        except toolkit.ObjectNotFound:
            raise toolkit.ValidationError(
                'Agent ({0}) does not exist.'.format(agent_id)
            )
    new_affiliation = AgentAffiliationQuery.create(**original_data_dict)
    if new_affiliation is None:
        raise toolkit.ValidationError(
            'Unable to create affiliation. Check the fields are valid.'
        )
    return new_affiliation.as_dict()


@action(schema.agent_create, help.agent_create)
def agent_create(original_data_dict):
    AgentQuery.validate(original_data_dict)
    new_agent = AgentQuery.create(**original_data_dict)
    if new_agent is None:
        raise toolkit.ValidationError(
            'Unable to create agent. Check the fields are valid.'
        )
    return new_agent.as_dict()


@action(schema.contribution_activity_create, help.contribution_activity_create)
def contribution_activity_create(context, original_data_dict, package_id, agent_id):
    try:
        toolkit.get_action('package_show')(context, {'id': package_id})
    except toolkit.ObjectNotFound:
        raise toolkit.ValidationError(
            'Cannot create activity for a package ({0}) that does not exist.'.format(
                package_id
            )
        )
    try:
        toolkit.get_action('agent_show')(context, {'id': agent_id})
    except toolkit.ObjectNotFound:
        raise toolkit.ValidationError(
            'Cannot create activity for an agent ({0}) that does not exist.'.format(
                agent_id
            )
        )
    new_activity = ContributionActivityQuery.create(**original_data_dict)
    PackageContributionActivityQuery.create(
        package_id=package_id, contribution_activity_id=new_activity.id
    )
    AgentContributionActivityQuery.create(
        agent_id=agent_id, contribution_activity_id=new_activity.id
    )
    return new_activity.as_dict()


@basic_action
@toolkit.chained_action
def package_create(next_func, context, data_dict):
    data_dict['author'] = 'pending'
    # we need the package ID to create links, but that's not created yet - so run the other
    # functions first
    created_pkg = next_func(context, data_dict)
    created_pkg['attribution'] = data_dict.get('attribution', '{}')
    parse_contributors(context, created_pkg)

    data_dict['author'] = get_author_string(package_id=created_pkg['id'])
    return data_dict



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

import re
from types import SimpleNamespace

import click


def multi_choice(question, options, default=0):
    """
    Provide the user with a list of options.

    :return: index of the chosen option
    """
    click.echo(question)
    for i, o in enumerate(options):
        click.echo('\t({0}) {1}'.format(i + 1, o))
    answer = click.prompt('Choose an option', default=default + 1)
    try:
        answer = int(answer)
        click.echo('')
        return answer - 1
    except:
        click.echo('That wasn\'t an option.', err=True)
        return multi_choice(question, options, default)


rgx = SimpleNamespace(
    opening=re.compile(r'\('),
    closing=re.compile(r'\)'),
    initials=re.compile(r'([A-Z])\.\s*'),
    name=re.compile(r'(?:(?:,\s?)|^|and )([A-Za-z]+, [^,(]+)'),
    reversed_name=re.compile(r'([A-Z][A-Za-z\s-]+)\s+([A-Z.\s]+)(?![a-z])(?:,\s+)?'),
    has_affiliation=re.compile(r'^([^()]+)\((.+)\)$'),
    no_affiliation=re.compile(r'^([^(),;]+)$'),
    initialism=re.compile(r'^[A-Z.]+$'),
    abbr=re.compile(r'([A-Z]|(?<=[^A-Za-z])[a-z])'),
)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.plugins import SingletonPlugin, implements, interfaces, toolkit
from ckanext.attribution.lib import helpers
from ckanext.attribution.model import (
    agent,
    agent_affiliation,
    agent_contribution_activity,
    contribution_activity,
    package_contribution_activity,
    relationships,
)
from ckanext.attribution.commands import cli
from ckanext.attribution import routes
from ckantools.loaders import create_actions, create_auth

try:
    from ckanext.doi.interfaces import IDoi

    doi_available = True
except ImportError:
    doi_available = False


class AttributionPlugin(SingletonPlugin):
    """
    A CKAN extension that adds support for complex attribution.
    """

    implements(interfaces.IActions, inherit=True)
    implements(interfaces.IAuthFunctions, inherit=True)
    implements(interfaces.IBlueprint, inherit=True)
    implements(interfaces.IClick)
    implements(interfaces.IConfigurable)
    implements(interfaces.IConfigurer)
    implements(interfaces.IFacets, inherit=True)
    implements(interfaces.IPackageController, inherit=True)
    implements(interfaces.ITemplateHelpers)
    if doi_available:
        implements(IDoi, inherit=True)

    # IActions
    def get_actions(self):
        from ckanext.attribution.logic.actions import (
            create,
            show,
            update,
            delete,
            extra,
        )

        return create_actions(create, show, update, delete, extra)

    # IAuthFunctions
    def get_auth_functions(self):
        from ckanext.attribution.logic.auth import create, show, update, delete, extra

        return create_auth(create, show, update, delete, extra)

    # IBlueprint
    def get_blueprint(self):
        return routes.blueprints

    # IClick
    def get_commands(self):
        return cli.get_commands()

    # IConfigurable
    def configure(self, config):
        contribution_activity.check_for_table()
        agent.check_for_table()
        agent_affiliation.check_for_table()
        agent_contribution_activity.check_for_table()
        package_contribution_activity.check_for_table()
        relationships.setup_relationships()

    # IConfigurer
    def update_config(self, config):
        toolkit.add_template_directory(config, 'theme/templates')
        toolkit.add_resource('theme/assets', 'ckanext-attribution')

    # IFacets
    def dataset_facets(self, facets_dict, package_type):
        enable_faceting = toolkit.asbool(
            toolkit.config.get('ckanext.attribution.enable_faceting', False)
        )
        if enable_faceting:
            facets_dict['author'] = toolkit._('Contributors')
        return facets_dict

    # IPackageController
    def before_index(self, pkg_dict):
        enable_faceting = toolkit.asbool(
            toolkit.config.get('ckanext.attribution.enable_faceting', False)
        )
        if enable_faceting:
            contributions = toolkit.get_action('package_contributions_show')(
                {}, {'id': pkg_dict['id']}
            )
            agents = [c['agent'] for c in contributions['contributions']]
            pkg_dict['author'] = [a['display_name'] for a in agents]
        return pkg_dict

    # ITemplateHelpers
    def get_helpers(self):
        return {
            'get_contributions': helpers.get_contributions,
            'can_edit': helpers.can_edit,
            'split_caps': helpers.split_caps,
            'get_cited_contributors': helpers.get_cited_contributors,
            'controlled_list': helpers.controlled_list,
            'doi_plugin': helpers.doi_plugin,
            'agent_from_user': helpers.agent_from_user,
            'user_contributions': helpers.user_contributions,
        }

    # IDoi
    def build_metadata_dict(self, pkg_dict, metadata_dict, errors):
        # Adds cited contributors as 'creators' and uncited contributors as 'contributors'.
        # 'creators' do not get roles/contributor types; 'contributors' do.
        all_contributors = helpers.get_cited_contributors(pkg_dict['id'])
        id_schemes = helpers.controlled_list('agent_external_id_schemes')

        def _make_contrib_dict(entry):
            d = {
                'is_org': entry['agent'].agent_type == 'org',
                'affiliations': [
                    a['agent'].display_name
                    for a in entry['agent'].package_affiliations(pkg_dict['id'])
                ],
            }
            if entry['agent'].agent_type == 'person':
                d['family_name'] = entry['agent'].family_name
                d['given_name'] = entry['agent'].given_names
            else:
                d['full_name'] = entry['agent'].name
            if entry['agent'].external_id and entry['agent'].external_id_scheme:
                scheme = entry['agent'].external_id_scheme
                scheme_label = id_schemes.get(scheme, {}).get('label', scheme)
                scheme_uri = id_schemes.get(scheme, {}).get('scheme_uri')
                d['identifiers'] = [
                    {
                        'identifier': entry['agent'].external_id_url,
                        'scheme': scheme_label,
                        'scheme_uri': scheme_uri,
                    }
                ]
            return d

        creators = []
        for c in all_contributors['cited']:
            creator_dict = _make_contrib_dict(c)
            creators.append(creator_dict)

        if len(creators) == 0:
            default_author = toolkit.config.get(
                'ckanext.doi.publisher',
                toolkit.config.get('ckan.site_title', 'Anonymous'),
            )
            creators.append({'full_name': default_author, 'is_org': True})

        contributors = []
        for c in all_contributors['uncited']:
            # Only one type can be sent to datacite, so use the datacite role/activity where the
            # contributor is ranked highest, e.g. if they're 2nd Editor and 1st DataManager,
            # DataManager is used. If none are ranked (or there are multiple with the same rank),
            # use the first alphabetically. If there are no datacite roles/activities for this
            # contributor, use 'Other' (because contributor type is a required field).
            contrib_dict = _make_contrib_dict(c)
            datacite_roles = sorted(
                [a for a in c['contributions'] if a.scheme == 'datacite'],
                key=lambda x: (x.order or len(c['contributions']), x.activity),
            )
            if datacite_roles:
                contrib_dict['contributor_type'] = datacite_roles[0].activity
            else:
                contrib_dict['contributor_type'] = 'Other'
            contributors.append(contrib_dict)

        metadata_dict['creators'] = creators
        metadata_dict['contributors'] = contributors

        return metadata_dict, errors



import itertools
import json

from ckan.plugins import toolkit
from ckanext.attribution.model.crud import (
    AgentQuery,
    AgentContributionActivityQuery,
    ContributionActivityQuery,
    AgentAffiliationQuery,
    PackageQuery,
    PackageContributionActivityQuery,
)


def split_list_by_action(input_list, crud_model, id_field='id'):
    result = {'new': [], 'updated': [], 'deleted': []}
    for item in input_list:
        meta = item.get('meta', {})
        if meta.get('is_editing', False):
            # ignore anything where the changes haven't been saved
            continue
        is_saved_edit = meta.get('is_saved_edit', False)
        is_hidden = meta.get('is_hidden', False)
        to_delete = meta.get('to_delete', False)
        is_active = (is_saved_edit or not is_hidden) and not to_delete

        is_new = meta.get('is_new', False)
        if is_active and is_new and item.get(id_field) is not None:
            # double check it's actually new - wrapped in an if so we don't have to check the
            # database for every single record
            try:
                crud_model.read(item.get(id_field))
                is_new = False
            except toolkit.ObjectNotFound:
                is_new = True

        if is_active and is_new:
            result['new'].append(item)
        elif is_active and meta.get('is_dirty', False):
            result['updated'].append(item)
        elif to_delete:
            result['deleted'].append(item)
    return result


def parse_contributors(context, data_dict):
    # if the context parameter is passed to any of the actions retrieved below with get_action,
    # they break. I don't know why, they just do.
    contributors = json.loads(data_dict.get('attribution', '{}'))

    try:
        pkg = PackageQuery.read(data_dict.get('id'))
    except:
        pkgs = PackageQuery.search(PackageQuery.m.name == data_dict.get('id'))
        pkg = pkgs[0] if len(pkgs) > 0 else None

    if pkg is None:
        raise toolkit.ObjectNotFound('This package does not exist.')
    pkg_id = pkg.id

    # agents
    agents = split_list_by_action(contributors.get('agents', []), AgentQuery)
    agent_cre = toolkit.get_action('agent_create')
    agent_upd = toolkit.get_action('agent_update')
    new_agents = {}
    for agent in agents['new']:
        gen_id = agent['id']
        del agent['id']
        new_id = agent_cre({}, agent)['id']
        new_agents[gen_id] = new_id
    for agent in agents['updated']:
        agent_upd({}, agent)
    # agents marked 'to_delete' almost certainly should not be deleted - only their activities
    # should be removed
    deleted_agents = [a['id'] for a in agents['deleted']]

    # activities
    activities = split_list_by_action(
        contributors.get('activities', []), ContributionActivityQuery
    )
    for agent in deleted_agents:
        activities['deleted'] += [
            r.as_dict()
            for r in AgentContributionActivityQuery.read_agent_package(agent, pkg_id)
        ]
    activities['deleted'] = list(set([a['id'] for a in activities['deleted']]))
    activity_cre = toolkit.get_action('contribution_activity_create')
    activity_upd = toolkit.get_action('contribution_activity_update')
    activity_del = toolkit.get_action('contribution_activity_delete')
    for activity in activities['new']:
        if activity['agent_id'] in deleted_agents:
            continue
        del activity['id']
        new_agent_id = new_agents.get(activity['agent_id'])
        if new_agent_id:
            activity['agent_id'] = new_agent_id
        activity['package_id'] = activity.get('package_id', pkg_id)
        activity_cre({}, activity)
    for activity in activities['updated']:
        if activity['agent_id'] in deleted_agents:
            continue
        activity_upd({}, activity)
    for activity in activities['deleted']:
        activity_del({}, {'id': activity})

    # citations (specialised activities)
    citations = split_list_by_action(
        contributors.get('citations', []), ContributionActivityQuery
    )
    for citation in citations['new']:
        if citation['agent_id'] in deleted_agents:
            continue
        del citation['id']
        citation['activity'] = '[citation]'
        citation['scheme'] = 'internal'
        new_agent_id = new_agents.get(citation['agent_id'])
        if new_agent_id:
            citation['agent_id'] = new_agent_id
        citation['package_id'] = citation.get('package_id', pkg_id)
        new_citation = activity_cre({}, citation)
    for citation in citations['updated']:
        if citation['agent_id'] in deleted_agents:
            continue
        updated_citation = activity_upd({}, citation)
    for citation in citations['deleted']:
        activity_del({}, {'id': citation['id']})
    # make sure the order is right
    all_citations = sorted(
        [
            c
            for c in PackageQuery.get_contributions(pkg_id)
            if c.activity == '[citation]'
        ],
        key=lambda x: x.order,
    )
    for i, c in enumerate(all_citations):
        if c.order != i + 1:
            activity_upd({}, {'id': c.id, 'order': i + 1})

    # affiliations
    affiliations = split_list_by_action(
        contributors.get('affiliations', []), AgentAffiliationQuery, 'db_id'
    )

    def affiliation_key(x):
        return sorted((x['agent_id'], x['other_agent_id']))

    affiliations = {
        gk: [
            list(a)[0]
            for k, a in itertools.groupby(
                sorted(gv, key=affiliation_key), key=affiliation_key
            )
        ]
        for gk, gv in affiliations.items()
    }
    affiliation_cre = toolkit.get_action('agent_affiliation_create')
    affiliation_upd = toolkit.get_action('agent_affiliation_update')
    affiliation_del = toolkit.get_action('agent_affiliation_delete')

    for aff in affiliations['new']:
        del aff['id']
        new_agent_id = new_agents.get(aff['agent_id'])
        aff['agent_a_id'] = new_agent_id or aff['agent_id']
        new_other_agent_id = new_agents.get(aff['other_agent_id'])
        aff['agent_b_id'] = new_other_agent_id or aff['other_agent_id']
        aff['package_id'] = pkg_id
        if aff['agent_a_id'] in deleted_agents or aff['agent_b_id'] in deleted_agents:
            continue
        affiliation_cre({}, aff)

    for aff in affiliations['updated']:
        aff['id'] = aff['db_id']
        new_agent_id = new_agents.get(aff['agent_id'])
        aff['agent_a_id'] = new_agent_id or aff['agent_id']
        new_other_agent_id = new_agents.get(aff['other_agent_id'])
        aff['agent_b_id'] = new_other_agent_id or aff['other_agent_id']
        if aff['agent_a_id'] in deleted_agents or aff['agent_b_id'] in deleted_agents:
            continue
        affiliation_upd({}, aff)

    for aff in affiliations['deleted']:
        affiliation_del({}, {'id': aff['db_id']})

    for agent in deleted_agents:
        for aff in AgentAffiliationQuery.read_agent(agent, pkg_id):
            affiliation_del({}, {'id': aff.id})


def get_author_string(package_id=None, citation_ids=None):
    if package_id is not None:
        citations = sorted(
            [
                c
                for c in PackageQuery.get_contributions(package_id)
                if c.activity == '[citation]'
            ],
            key=lambda x: x.order,
        )
    elif citation_ids is not None:
        citations = sorted(
            [ContributionActivityQuery.read(c) for c in citation_ids],
            key=lambda x: x.order,
        )
    else:
        citations = []

    if len(citations) == 0:
        return toolkit.config.get(
            'ckanext.doi.publisher', toolkit.config.get('ckan.site_title', 'Anonymous')
        )
    else:
        return '; '.join([c.agent.citation_name for c in citations])



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckanext.attribution.model.agent_affiliation import (
    AgentAffiliation,
    agent_affiliation_table,
)

from ._base import BaseQuery
from sqlalchemy import or_, and_


class AgentAffiliationQuery(BaseQuery):
    """
    
    """

    # model and table (subclasses should override)
    m = AgentAffiliation
    t = agent_affiliation_table

    @classmethod
    def read_agent(cls, agent_id, package_id=None):
        filters = [or_(cls.m.agent_a_id == agent_id, cls.m.agent_b_id == agent_id)]
        if package_id is not None:
            filters.append(cls.m.package_id == package_id)
        return cls.search(and_(*filters))



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-dataset-contributors
# Created by the Natural History Museum in London, UK

import click
from ckan.model import Session
from ckan.model.package_extra import PackageExtra
from ckan.plugins import toolkit
from ckanext.attribution.commands import migration
from ckanext.attribution.logic.actions.helpers import get_author_string
from ckanext.attribution.model import (
    agent,
    agent_affiliation,
    agent_contribution_activity,
    contribution_activity,
    package_contribution_activity,
    relationships,
)
from ckanext.attribution.model.crud import (
    AgentQuery,
    PackageQuery,
    PackageContributionActivityQuery,
    AgentAffiliationQuery,
    AgentContributionActivityQuery,
)
from fuzzywuzzy import process, fuzz
from sqlalchemy import and_, or_


def get_commands():
    return [attribution]


@click.group()
def attribution():
    """
    Commands for the ckanext-attribution plugin.
    """
    pass


@attribution.command()
def initdb():
    """
    Create database tables required for this extension.
    """
    contribution_activity.check_for_table()
    agent.check_for_table()
    agent_affiliation.check_for_table()
    agent_contribution_activity.check_for_table()
    package_contribution_activity.check_for_table()
    relationships.setup_relationships()


@attribution.command()
@click.argument('ids', nargs=-1)
def sync(ids):
    """
    Pull updated details for agents from external services like ORCID and ROR.

    Only applies when an external_id has already been set.
    """
    agent_external_update = toolkit.get_action('agent_external_update')
    if not ids:
        ids = [a.id for a in AgentQuery.all() if a.external_id]
    click.echo('Attempting to sync {0} records.'.format(len(ids)))
    errors = []
    with click.progressbar(ids) as bar:
        for _id in bar:
            try:
                agent_external_update({'ignore_auth': True}, {'id': _id})
            except Exception as e:
                errors.append('Error ({0}): {1}'.format(_id, e))
    failed = len(errors)
    total = len(ids)
    click.echo('Updated {0}/{1} ({2} failed)'.format(total - failed, total, failed))
    for e in errors:
        click.echo(e, err=True)


@attribution.command()
@click.argument('ids', nargs=-1)
def refresh_packages(ids):
    """
    Update the author string for all (or the specified) packages.
    """
    if not ids:
        ids = list(set([r.package_id for r in PackageContributionActivityQuery.all()]))
    click.echo(
        'Attempting to update the author field for {0} packages.'.format(len(ids))
    )
    errors = []
    with click.progressbar(ids) as bar:
        for _id in bar:
            try:
                authors = get_author_string(package_id=_id)
                PackageQuery.update(_id, author=authors)
            except Exception as e:
                errors.append('Error ({0}): {1}'.format(_id, e))
    failed = len(errors)
    total = len(ids)
    click.echo('Updated {0}/{1} ({2} failed)'.format(total - failed, total, failed))
    for e in errors:
        click.echo(e, err=True)


@attribution.command()
@click.argument('ids', nargs=-1)
@click.option(
    '--limit', help='Process n packages at a time (best for testing/debugging).'
)
def agent_external_search(ids, limit):
    if ids:
        agents = AgentQuery.search(AgentQuery.m.id.in_(ids))
    else:
        agents = AgentQuery.search(AgentQuery.m.external_id.is_(None))
    agents = sorted(agents, key=lambda x: -len(x.contribution_activities))
    if limit:
        agents = agents[: int(limit)]
    total = len(agents)
    click.echo(f'{total} contributors without external IDs found.')
    updater = migration.APISearch()
    for i, a in enumerate(agents):
        click.echo(f'Searching {i} of {total}')
        update_dict = updater.update(a.as_dict())
        AgentQuery.update(a.id, **update_dict)


@attribution.command()
@click.option('--q')
@click.option('--match-threshold', default=75)
def merge_agents(q, match_threshold):
    agents = toolkit.get_action('agent_list')({}, {'q': q})
    all_agents = AgentQuery.all()
    merging = []
    for a in agents:
        if a['id'] in merging:
            continue
        other_agents = [o.display_name for o in all_agents if o.id != a['id']]
        compare = process.extract(
            a['display_name'], other_agents, limit=10, scorer=fuzz.token_set_ratio
        )
        matches = [m for m in compare if m[1] >= int(match_threshold)]
        to_merge = []
        while matches:
            ix = migration.multi_choice(
                f'Choose a{"nother" if len(to_merge) > 0 else ""} record to merge with {a["display_name"]}:',
                [f'{m}: {x}%' for m, x in matches] + ['None of these'],
                default=len(matches),
            )
            if ix == len(matches):
                break
            match_name = matches.pop(ix)[0]
            to_merge.append(
                next(
                    r
                    for r in all_agents
                    if r.display_name == match_name and r.id != a['id']
                )
            )
        if len(to_merge) == 0:
            continue
        to_merge.append(AgentQuery.read(a['id']))
        merging += [m.id for m in to_merge]
        has_external_id = [r for r in to_merge if r.external_id is not None]
        if len(has_external_id) == 1:
            base_record = has_external_id[0]
        else:
            choices = has_external_id if len(has_external_id) > 1 else to_merge
            ix = migration.multi_choice(
                'Choose base record:',
                [f'{r.display_name} ({r.external_id_url or "no ID"})' for r in choices],
            )
            base_record = choices[ix]
        not_base_record = [r for r in to_merge if r.id != base_record.id]
        for merging_record in not_base_record:
            # update affiliations
            for aff in merging_record.affiliations:
                if aff['agent'].id == base_record.id:
                    AgentAffiliationQuery.delete(aff['affiliation'].id)
                else:
                    k = (
                        'agent_a_id'
                        if aff['affiliation'].agent_a_id == merging_record.id
                        else 'agent_b_id'
                    )
                    AgentAffiliationQuery.update(
                        aff['affiliation'].id, **{k: base_record.id}
                    )
            base_record_activities = AgentContributionActivityQuery.search(
                AgentContributionActivityQuery.m.agent_id == base_record.id
            )
            merging_activities = AgentContributionActivityQuery.search(
                AgentContributionActivityQuery.m.agent_id == merging_record.id
            )
            for activity in merging_activities:
                name = activity.contribution_activity.activity
                pkg = activity.contribution_activity.package.id
                scheme = activity.contribution_activity.scheme
                matching_activities = [
                    x
                    for x in base_record_activities
                    if (
                        x.contribution_activity.activity == name
                        and x.contribution_activity.package.id == pkg
                        and x.contribution_activity.scheme == scheme
                    )
                ]
                if len(matching_activities) == 0:
                    AgentContributionActivityQuery.update(
                        activity.id, agent_id=base_record.id
                    )
                else:
                    matching_activities.append(activity)
                    matching_activities = sorted(
                        matching_activities, key=lambda x: x.contribution_activity.order
                    )
                    first_activity = matching_activities.pop(0)
                    AgentContributionActivityQuery.update(
                        first_activity.id, agent_id=base_record.id
                    )
                    for other_activity in matching_activities:
                        AgentContributionActivityQuery.delete(other_activity.id)
            AgentQuery.delete(merging_record.id)
            click.echo(
                f'Merged {merging_record.display_name} ({merging_record.id}) into {base_record.display_name} ({base_record.id}).'
            )
    if len(merging) == 0:
        click.echo('Nothing to merge.')


@attribution.command()
@click.option(
    '--limit', help='Process n packages at a time (best for testing/debugging).'
)
@click.option('--dry-run', help='Don\'t save anything to the database.', is_flag=True)
@click.option(
    '--search-api/--no-search-api',
    help='Search external APIs (e.g. ORCID) for details.',
    default=True,
)
def migratedb(limit, dry_run, search_api):
    """
    Semi-manual migration script that attempts to extract individual contributors from
    'author' and 'contributor' fields (if present) in order to create Agent and
    ContributionActivity records for them.
    """
    if not dry_run:
        click.secho(
            'Attempting to migrate contributors. It is HIGHLY recommended that you back up your '
            'database before running this.',
            fg='red',
        )
        click.confirm('Continue?', default=False, abort=True)
    converted_packages = [r.package_id for r in PackageContributionActivityQuery.all()]
    unconverted_packages = PackageQuery.search(
        ~PackageQuery.m.id.in_(converted_packages)
    )
    contribution_extras = {
        p.id: Session.query(PackageExtra)
        .filter(PackageExtra.package_id == p.id, PackageExtra.key == 'contributors')
        .first()
        for p in unconverted_packages
    }
    total = len(unconverted_packages)
    limit = int(limit or total)
    parser = migration.Parser()

    for i, pkg in enumerate(unconverted_packages[:limit]):
        click.echo('Processing package {0} of {1}.\n'.format(i + 1, total))
        parser.run(pkg.author, pkg.id, 'author')
        if contribution_extras.get(pkg.id) is not None:
            extras = contribution_extras.get(pkg.id).value
            if not isinstance(extras, str):
                parser.run(extras, pkg.id, 'contributor')

    combiner = migration.Combiner(parser)
    combined = combiner.run()
    if search_api:
        api_updater = migration.APISearch()
        for agnt in combined:
            api_updater.update(agnt)
    click.echo(f'\n\n{len(combined)} contributors found.')
    if dry_run:
        click.echo('Exiting before saving to the database.')
        return
    agent_lookup = {}
    agent_create = toolkit.get_action('agent_create')
    contribution_activity_create = toolkit.get_action('contribution_activity_create')
    agent_affiliation_create = toolkit.get_action('agent_affiliation_create')
    remove_keys = ['packages', 'affiliations', 'key', 'all_names']
    for a in combined:
        try:
            # create the agent (check it doesn't exist first)
            agent_dict = {**{k: v for k, v in a.items() if k not in remove_keys}}
            if a['agent_type'] == 'person':
                filters = [
                    and_(
                        AgentQuery.m.family_name == a['family_name'],
                        AgentQuery.m.given_names == a['given_names'],
                    )
                ]
            else:
                filters = [AgentQuery.m.name == a['name']]
            if a.get('external_id'):
                filters.append(AgentQuery.m.external_id == a.get('external_id'))
            matches = AgentQuery.search(or_(*filters))
            if len(matches) == 1:
                new_agent = matches[0].as_dict()
                click.echo(f'MATCHED "{a["key"]}"')
            elif len(matches) > 1:
                choice_ix = migration.multi_choice(
                    f'Does "{a["key"]}" match any of these existing agents?',
                    [m.display_name for m in matches] + ['None of these'],
                )
                if choice_ix == len(matches):
                    del a['external_id']
                    del a['external_id_scheme']
                    new_agent = agent_create({'ignore_auth': True}, agent_dict)
                    click.echo(f'CREATED "{a["key"]}"')
                else:
                    new_agent = matches[choice_ix].as_dict()
                    click.echo(f'MATCHED "{a["key"]}"')
            else:
                new_agent = agent_create({'ignore_auth': True}, agent_dict)
                click.echo(f'CREATED "{a["key"]}"')
            agent_lookup[a['key']] = new_agent['id']
            # then activities
            for pkg, order in a['packages'].get('author', []):
                # create citation
                contribution_activity_create(
                    {'ignore_auth': True},
                    {
                        'activity': '[citation]',
                        'scheme': 'internal',
                        'order': order,
                        'package_id': pkg,
                        'agent_id': new_agent['id'],
                    },
                )
                # then the actual activity
                contribution_activity_create(
                    {'ignore_auth': True},
                    {
                        'activity': 'Unspecified',
                        'scheme': 'internal',
                        'package_id': pkg,
                        'agent_id': new_agent['id'],
                    },
                )
            for pkg, _ in a['packages'].get('contributor', []):
                # just the activity for this one
                contribution_activity_create(
                    {'ignore_auth': True},
                    {
                        'activity': 'Unspecified',
                        'scheme': 'internal',
                        'package_id': pkg,
                        'agent_id': new_agent['id'],
                    },
                )
        except Exception as e:
            # very broad catch just so it doesn't ruin everything if one thing breaks
            click.echo(f'Skipping {a["key"]} due to error: {e}', err=True)
    # finally, the affiliations
    for pkg, pairs in combiner.affiliations.items():
        for agent_a, agent_b in pairs:
            try:
                agent_affiliation_create(
                    {'ignore_auth': True},
                    {
                        'agent_a_id': agent_lookup[agent_a],
                        'agent_b_id': agent_lookup[agent_b],
                        'package_id': pkg,
                    },
                )
            except Exception as e:
                # very broad catch just so it doesn't ruin everything if one thing breaks
                click.echo(
                    f'Skipping {agent_a} + {agent_b} affiliation due to error: {e}',
                    err=True,
                )

    # finally finally, update the package author strings
    for pkg in unconverted_packages[:limit]:
        try:
            authors = get_author_string(package_id=pkg.id)
            PackageQuery.update(pkg.id, author=authors)
        except Exception as e:
            # very broad catch just so it doesn't ruin everything if one thing breaks
            click.echo(f'Skipping {pkg.id} due to error: {e}', err=True)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckantools.validators import list_of_strings

from ckan.plugins import toolkit

# grab all the validator functions upfront
ignore_missing = toolkit.get_validator('ignore_missing')
not_missing = toolkit.get_validator('not_missing')
boolean_validator = toolkit.get_validator('boolean_validator')
isodate_validator = toolkit.get_validator('isodate')
int_validator = toolkit.get_validator('int_validator')

# CREATE ===========================================================================================

agent_affiliation_create = {
    'agent_a_id': [not_missing, str],
    'agent_b_id': [not_missing, str],
    'affiliation_type': [ignore_missing, str],
    'description': [ignore_missing, str],
    'start_date': [ignore_missing, isodate_validator],
    'end_date': [ignore_missing, isodate_validator],
}

agent_create = {
    'agent_type': [not_missing, str],
    'family_name': [ignore_missing, str],
    'given_names': [ignore_missing, str],
    'given_names_first': [ignore_missing, boolean_validator],
    'user_id': [ignore_missing, str],
    'name': [ignore_missing, str],
}

contribution_activity_create = {
    'package_id': [not_missing, str],
    'agent_id': [not_missing, str],
    'activity': [not_missing, str],
    'scheme': [not_missing, str],
    'level': [ignore_missing, str],
    'time': [ignore_missing, isodate_validator],
}

# DELETE ===========================================================================================

agent_affiliation_delete = {'id': [not_missing, str]}

agent_delete = {'id': [not_missing, str]}

agent_contribution_activity_delete = {'id': [not_missing, str]}

contribution_activity_delete = {
    'id': [not_missing, str],
}

package_contribution_activity_delete = {'id': [not_missing, str]}

# EXTRA ============================================================================================

attribution_controlled_lists = {'lists': [ignore_missing, list_of_strings()]}

agent_external_search = {
    'q': [not_missing, str],
    'sources': [ignore_missing, list_of_strings()],
}

agent_external_read = {
    'agent_id': [ignore_missing, str],
    'external_id': [ignore_missing, str],
    'external_id_scheme': [ignore_missing, str],
    'diff': [ignore_missing, boolean_validator],
}

validate_external_id = {
    'external_id': [not_missing, str],
    'external_id_scheme': [not_missing, str],
}

# SHOW =============================================================================================

agent_affiliation_show = {'id': [not_missing, str]}

agent_show = {'id': [not_missing, str]}

agent_list = {'q': [ignore_missing, str], 'mode': [ignore_missing, str]}

agent_contribution_activity_show = {'id': [not_missing, str]}

contribution_activity_show = {'id': [not_missing, str]}

package_contribution_activity_show = {'id': [not_missing, str]}

package_contributions_show = {
    'id': [not_missing, str],
    'limit': [ignore_missing, int_validator],
    'offset': [ignore_missing, int_validator],
}

agent_affiliations = {
    'agent_id': [not_missing, str],
    'package_id': [ignore_missing, str],
}

# UPDATE ===========================================================================================

agent_affiliation_update = {
    'id': [not_missing, str],
    'agent_a_id': [ignore_missing, str],
    'agent_b_id': [ignore_missing, str],
    'affiliation_type': [ignore_missing, str],
    'description': [ignore_missing, str],
    'start_date': [ignore_missing, isodate_validator],
    'end_date': [ignore_missing, isodate_validator],
}

agent_update = {
    'id': [not_missing, str],
    'agent_type': [ignore_missing, str],
    'family_name': [ignore_missing, str],
    'given_names': [ignore_missing, str],
    'given_names_first': [ignore_missing, boolean_validator],
    'user_id': [ignore_missing, str],
    'name': [ignore_missing, str],
}

agent_external_update = {'id': [not_missing, str]}

contribution_activity_update = {
    'id': [not_missing, str],
    'activity': [ignore_missing, str],
    'scheme': [ignore_missing, str],
    'level': [ignore_missing, str],
    'time': [ignore_missing, isodate_validator],
}



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

"""
These relationships are defined separately to the object/table declarations to allow
them to reference each other without circular imports.
"""

from ckan.model import Package, User, meta
from sqlalchemy import or_
from sqlalchemy.orm import backref, relationship, attributes

from .agent import Agent, agent_table
from .agent_affiliation import AgentAffiliation, agent_affiliation_table
from .agent_contribution_activity import (
    AgentContributionActivity,
    agent_contribution_activity_table,
)
from .contribution_activity import ContributionActivity, contribution_activity_table
from .package_contribution_activity import (
    PackageContributionActivity,
    package_contribution_activity_table,
)


def setup_relationships():
    """
    Defines relationships/joins between the new models in ckanext-attribution.
    """

    # Agent
    manager = attributes.manager_of_class(Agent)
    if not manager or not manager.is_mapped:
        meta.mapper(
            Agent,
            agent_table,
            properties={
                'user': relationship(
                    User,
                    backref=backref('agent', cascade='all, delete-orphan'),
                    primaryjoin=agent_table.c.user_id.__eq__(User.id),
                ),
                'contribution_activities': relationship(
                    ContributionActivity, secondary=agent_contribution_activity_table
                ),
            },
        )

    # ContributionActivity
    manager = attributes.manager_of_class(ContributionActivity)
    if not manager or not manager.is_mapped:
        meta.mapper(
            ContributionActivity,
            contribution_activity_table,
            properties={
                'agent': relationship(
                    Agent, secondary=agent_contribution_activity_table, uselist=False
                ),
                'package': relationship(
                    Package,
                    secondary=package_contribution_activity_table,
                    uselist=False,
                ),
            },
        )

    # AgentAffiliation
    manager = attributes.manager_of_class(AgentAffiliation)
    if not manager or not manager.is_mapped:
        meta.mapper(
            AgentAffiliation,
            agent_affiliation_table,
            properties={
                '_agent_backref': relationship(
                    Agent,
                    backref=backref('_affiliations', cascade='all, delete-orphan'),
                    primaryjoin=or_(
                        agent_affiliation_table.c.agent_a_id.__eq__(Agent.id),
                        agent_affiliation_table.c.agent_b_id.__eq__(Agent.id),
                    ),
                ),
                '_agent_a': relationship(
                    Agent,
                    primaryjoin=agent_affiliation_table.c.agent_a_id.__eq__(Agent.id),
                ),
                '_agent_b': relationship(
                    Agent,
                    primaryjoin=agent_affiliation_table.c.agent_b_id.__eq__(Agent.id),
                ),
                'package': relationship(
                    Package,
                    primaryjoin=agent_affiliation_table.c.package_id == Package.id,
                ),
            },
        )

    # AgentContributionActivity
    manager = attributes.manager_of_class(AgentContributionActivity)
    if not manager or not manager.is_mapped:
        meta.mapper(
            AgentContributionActivity,
            agent_contribution_activity_table,
            properties={
                'agent': relationship(
                    Agent,
                    backref=backref(
                        'contribution_activity_link', cascade='all, delete-orphan'
                    ),
                    primaryjoin=agent_contribution_activity_table.c.agent_id.__eq__(
                        Agent.id
                    ),
                ),
                'contribution_activity': relationship(
                    ContributionActivity,
                    backref=backref('agent_link', cascade='all, delete-orphan'),
                    primaryjoin=agent_contribution_activity_table.c.contribution_activity_id.__eq__(
                        ContributionActivity.id
                    ),
                ),
            },
        )

    # PackageContributionActivity
    manager = attributes.manager_of_class(PackageContributionActivity)
    if not manager or not manager.is_mapped:
        meta.mapper(
            PackageContributionActivity,
            package_contribution_activity_table,
            properties={
                'package': relationship(
                    Package,
                    backref=backref(
                        'contribution_activity_link', cascade='all, delete-orphan'
                    ),
                    primaryjoin=package_contribution_activity_table.c.package_id.__eq__(
                        Package.id
                    ),
                ),
                'contribution_activity': relationship(
                    ContributionActivity,
                    backref=backref('package_link', cascade='all, delete-orphan'),
                    primaryjoin=package_contribution_activity_table.c.contribution_activity_id.__eq__(
                        ContributionActivity.id
                    ),
                ),
            },
        )



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.model import Session
from ckan.plugins import toolkit
from ckanext.attribution.lib.orcid_api import OrcidApi
from ckanext.attribution.lib.ror_api import RorApi
from ckanext.attribution.model.agent import Agent, agent_table
from requests import HTTPError

from ._base import BaseQuery


class AgentQuery(BaseQuery):
    """
    CRUD methods for :class:`~ckanext.attribution.model.agent.Agent`.

    Fields
    ======
    :param agent_type: broad type of agent; usually 'person' or 'org'
    :type agent_type: str
    :param external_id: the agent's ID from an external service like ORCID or ROR
    :type external_id: str, optional
    :param external_id_scheme: the name of the scheme for the external ID, e.g. 'orcid' or 'ror'
    :type external_id_scheme: str, optional
    :param family_name: family name of an person [person only, required]
    :type family_name: str, optional
    :param given_names: given name(s) or initials of an person [person only, required]
    :type given_names: str, optional
    :param given_names_first: whether given names should be displayed before the family name
                              (default True) [person only, optional]
    :type given_names_first: bool, optional
    :param user_id: the ID for a registered user of this CKAN instance associated with this agent
                    [person only, optional]
    :type user_id: str, optional
    :param name: name of an organisation [org only, required]
    :type name: str, optional
    """

    #: :type: The associated database model type.
    m = Agent

    #: :sqlalchemy.Table: The associated database table instance (agent_table).
    t = agent_table

    @classmethod
    def validate(cls, data_dict):
        data_dict = super(AgentQuery, cls).validate(data_dict)
        valid_agent_types = ['person', 'org', 'other']
        agent_type = toolkit.get_or_bust(data_dict, 'agent_type')
        if agent_type not in valid_agent_types:
            raise toolkit.Invalid(
                'Agent type must be one of {0}'.format(', '.join(valid_agent_types))
            )

        valid_params = {
            'person': dict(
                required=['family_name', 'given_names'], optional=['given_names_first']
            ),
            'org': dict(required=['name'], optional=['location']),
            'other': dict(required=[], optional=[]),
        }
        required = ['agent_type'] + valid_params[agent_type]['required']
        optional = ['user_id', 'external_id', 'external_id_scheme'] + valid_params[
            agent_type
        ]['optional']
        for k in required:
            if k not in data_dict:
                raise toolkit.ValidationError('{0} is a required field.'.format(k))
        if 'external_id' in data_dict and 'external_id_scheme' not in data_dict:
            raise toolkit.ValidationError(
                'external_id_scheme is a required field when external_id is set.'
            )
        all_fields = required + optional
        for k in data_dict:
            if k not in all_fields:
                data_dict[k] = None
        return data_dict

    @classmethod
    def read_external_id(cls, eid):
        """
        Retrieve an agent record by its external identifier.

        :param eid: the full external ID (e.g. ORCID or ROR ID) of the record
        :type eid: str
        :returns: One agent or None if not found.
        :rtype: Agent
        """
        return Session.query(Agent).filter(Agent.external_id == eid).first()

    @classmethod
    def read_from_external_api(cls, agent_id, api=None):
        record = cls.read(agent_id)
        if record.external_id is None:
            raise Exception(toolkit._('Record does not have an external ID set.'))
        if record.external_id_scheme is None:
            raise Exception(toolkit._('External ID scheme not set.'))
        external_id_schemes = toolkit.get_action('attribution_controlled_lists')(
            {}, {'lists': ['agent_external_id_schemes']}
        )['agent_external_id_schemes']
        if record.external_id_scheme not in external_id_schemes:
            raise Exception(
                toolkit._(
                    'External ID scheme "{0}" not recognised.'.format(
                        record.external_id_scheme
                    )
                )
            )
        if record.external_id_scheme == 'orcid':
            return cls._read_from_orcid_api(record, api)
        elif record.external_id_scheme == 'ror':
            return cls._read_from_ror_api(record)
        else:
            raise NotImplementedError

    @classmethod
    def _read_from_orcid_api(cls, record, api=None):
        """
        Update the data for an agent using the ORCID API and the agent's stored ORCID.

        :param record: the existing record
        :type record: Agent
        :param api: an API instance already in use (useful if performing this action over many
                    agent records, to avoid instantiating many API connections) (Default value = None)
        :type api: OrcidApi
        :returns: the updated agent record
        :rtype: Agent
        """
        if api is None:
            api = OrcidApi()
        try:
            orcid_record = api.as_agent_record(api.read(record.external_id))
        except HTTPError:
            orcid_results = api.search(orcid_q=record.external_id)
            if orcid_results['total'] == 0:
                raise Exception(
                    toolkit._(
                        'This ORCID ({0}) does not exist.'.format(record.external_id)
                    )
                )
            else:
                orcid_record = orcid_results['records'][0]
        orcid_record['id'] = record.id
        return cls.validate(orcid_record)

    @classmethod
    def _read_from_ror_api(cls, record):
        """
        Update the data for an agent using the ROR API and the agent's stored ROR ID.

        :param record: the existing record
        :type record: Agent
        :returns: the updated agent record
        :rtype: Agent
        """
        api = RorApi()
        ror_record = api.read(record.external_id)
        if ror_record is None:
            raise Exception(
                toolkit._(
                    'Unable to find this ROR ID ({0}).'.format(record.external_id)
                )
            )
        updated_entry = api.as_agent_record(ror_record)
        updated_entry['id'] = record.id
        return cls.validate(updated_entry)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from collections import Counter
from dataclasses import dataclass
from textwrap import shorten

import click
import spacy
from nameparser import HumanName
from prompt_toolkit import prompt

from .common import rgx, multi_choice


@dataclass
class ParsedSegment:
    name: str
    text: str
    affiliations: list
    packages: dict


class Parser(object):
    """
    Extracts names and affiliations from text.
    """

    def __init__(self):
        self.contributors = {'person': {}, 'org': {}, 'other': {}}
        self.affiliations = {}
        spacy_model = 'en_core_web_trf'
        try:
            self.nlp = spacy.load(spacy_model)
        except OSError:
            spacy.cli.download(spacy_model)
            self.nlp = spacy.load(spacy_model)

    def run(self, txt, pkg_id, contrib_type):
        """
        Run the whole process over a line of text.

        :param txt: the chunk of text to process
        :param pkg_id: associated package
        :param contrib_type: author, contributor, or affiliation
        :return: list of ParsedSegment instances extracted from the text
        """
        if not self.validate(txt):
            return
        segments = []
        for line in txt.split('\n'):
            line = line.replace('\\r', '')
            sublines = self.split(line)
            # bulk sort for large lists
            if len(sublines) > 20 and click.confirm(
                f'Do you want to set the same type for all {len(sublines)} contributors found in this block?'
            ):
                i = multi_choice(
                    f'What type of contributors are in "{shorten(line, width=50, placeholder="...")}"?',
                    self.contributors.keys(),
                )
                _type = list(self.contributors.keys())[i]
            else:
                _type = None
            for i, subline in enumerate(sublines):
                if subline == '':
                    continue
                if contrib_type != 'affiliation':
                    name, affiliations = self.extract_affiliations(subline)
                    for a in affiliations:
                        self.affiliations[pkg_id] = self.affiliations.get(
                            pkg_id, []
                        ) + [(name.strip(), a)]
                        parsed_affiliation = ParsedSegment(
                            a.strip(),
                            text=a,
                            affiliations=[],
                            packages={'affiliation': (pkg_id, None)},
                        )
                        self.sort_contributor(parsed_affiliation)
                else:
                    name = subline
                    affiliations = []
                order = (i + 1) if contrib_type == 'author' else None
                parsed_segment = ParsedSegment(
                    name=name.strip(),
                    text=subline,
                    affiliations=affiliations,
                    packages={contrib_type: (pkg_id, order)},
                )
                self.sort_contributor(parsed_segment, _type)
                segments.append(parsed_segment)
        return segments

    def validate(self, txt):
        """
        Check the text can/should actually be parsed.

        :return: True/False
        """
        if txt is None:
            return False
        parsed = self.nlp(txt)
        # check that it has some proper nouns and doesn't just look like a sentence
        tokens = [t for t in parsed]
        pos = Counter([t.pos_ for t in tokens])
        if len(tokens) == 0:
            return []
        pc_proper_nouns = pos.get('PROPN', 0) / len(tokens)
        if pc_proper_nouns < 0.5:
            click.echo('\nThis text doesn\'t look right:')
            click.echo(txt)
            return not click.confirm('Skip it?')
        return True

    def split(self, txt):
        """
        Uses multiple sub-methods to attempt to split the text into individual
        contributors.

        :param txt:
        :return:
        """
        lines = [ln.strip() for ln in txt.split('\n')]
        segments = [
            rgx.initials.sub('\\1 ', s).strip() for ln in lines for s in ln.split(';')
        ]
        names = []

        def _process_segment(segment):
            if ';' in segment:
                semicolon_splits = [s.strip() for s in segment.split(';')]
            else:
                semicolon_splits = []
            rgx_splits = rgx.name.findall(segment)
            rv_names = self._split_by_reversed_name(segment)
            nlp_ent_splits = self._split_by_nlp_ents(segment)
            options = []
            printable_options = []
            all_options = [
                nlp_ent_splits,
                rv_names,
                rgx_splits,
                semicolon_splits,
                [segment],
            ]
            all_options = sorted(all_options, key=lambda x: -len(''.join(x)))
            all_options = sorted(all_options, key=lambda x: -len(x))
            for o in all_options:
                if len(o) == 0:
                    continue
                printable = '; '.join(o) + ' ({0} fragments)'.format(len(o))
                if printable in printable_options:
                    continue
                options.append(o)
                printable_options.append(printable)
            options.append(None)
            printable_options.append('None of these')
            if len(options) > 1:
                click.echo(segment)
                ix = multi_choice('Which one looks right?', printable_options)
                return options[ix]

        if len(segments) > 1:
            click.echo(
                f'{len(segments)} found in "{shorten(txt, width=50, placeholder="...")}". e.g.:'
            )
            for s in segments[:5]:
                click.echo('\t' + s)
            if click.confirm('Skip individual processing of these segments?'):
                return segments

        for t in segments:
            splits = None
            while not splits:
                splits = _process_segment(t)
                if splits:
                    names += splits
                else:
                    t = prompt(
                        'Edit the line (try adding ";" between contributors): ',
                        default=t,
                    )
        return names

    def _split_by_nlp_sep(self, txt):
        """
        Finds entities using spacy, then attempts to identify the character(s)
        separating them and splits by that.
        """
        parsed = self.nlp(txt)
        if len(parsed.ents) > 2:
            start_char = parsed.ents[0].end_char
            end_char = parsed.ents[1].start_char
            sep = txt[start_char:end_char].strip()
            if sep == '':
                sep = '\n'
            return [s.strip() for s in txt.split(sep)]
        return [txt]

    def _split_by_nlp_ents(self, txt):
        """
        Extract all entities from the text using spacy.
        """
        parsed = self.nlp(txt)
        return [ent.text for ent in parsed.ents]

    def _split_by_reversed_name(self, txt):
        names = rgx.reversed_name.findall(txt)
        return [', '.join(n) for n in names]

    def extract_affiliations(self, txt):
        """
        Uses regexes to find probable affiliations in parentheses.

        :return: contributor name, list of affiliations
        """
        has_affiliation = rgx.has_affiliation.match(txt)
        no_affiliation = rgx.no_affiliation.match(txt)
        if has_affiliation is not None:
            return has_affiliation.groups()[0], self.split(has_affiliation.groups()[1])
        if no_affiliation is not None:
            return no_affiliation.groups()[0], []
        else:
            return txt, []

    def sort_contributor(self, c: ParsedSegment, default_type=None):
        """
        Sort a contributor into lists based on agent type.
        """
        name = HumanName(c.name)
        initials = ''.join(rgx.abbr.findall(c.name))
        _type = default_type
        if name.last in self.contributors['person']:
            _type = 'person'
        else:
            for k, v in self.contributors.items():
                if k == 'person':
                    continue
                if initials in v:
                    if any([x.name == c.name for x in v[initials]]):
                        _type = k
        if _type is None:
            i = multi_choice(
                'What type of contributor is "{0}"?'.format(c.name),
                self.contributors.keys(),
            )
            _type = list(self.contributors.keys())[i]
        if _type == 'person':
            if name.last == '':
                name.last = name.first
                name.first = '?'
            if name.first == '':
                name.first = '?'
            c.name = name
            family_name_records = self.contributors[_type].get(name.last, {})
            initial_records = family_name_records.get(name.first[0], []) + [c]
            family_name_records[name.first[0]] = initial_records
            self.contributors[_type][name.last] = family_name_records
        else:
            self.contributors[_type][initials] = self.contributors[_type].get(
                initials, []
            ) + [c]



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit
from flask import Blueprint
from ckanext.attribution.lib.helpers import user_contributions

blueprint = Blueprint(name='attribution_user', import_name=__name__)


@blueprint.route('/user/<username>/contributions', methods=['GET'])
def datasets(username):
    """
    Render a list of datasets that this user has contributed to.

    :param username: The username
    :return: str
    """
    try:
        toolkit.check_access('user_show', {}, {'id': username})
    except toolkit.NotAuthorized:
        toolkit.abort(403, toolkit._('Not authorized to see this page'))
    user = toolkit.get_action('user_show')(
        {'for_view': True}, {'id': username, 'include_num_followers': True}
    )
    return toolkit.render(
        'user/contributions.html',
        extra_vars=dict(contributions=user_contributions(user['id']), user_dict=user),
    )



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

import click
from ckanext.attribution.lib.orcid_api import OrcidApi
from ckanext.attribution.lib.ror_api import RorApi
from prompt_toolkit import prompt
from nameparser import HumanName

from .common import multi_choice


class APISearch(object):
    """
    Search for a contributor on an external API and download any relevant details.
    """

    def __init__(self):
        self.api = {'ORCID': OrcidApi(), 'ROR': RorApi()}

    def update(self, contrib):
        if contrib['agent_type'] == 'person':
            api_func = self.search_orcid
        elif contrib['agent_type'] == 'org':
            api_func = self.search_ror
        else:
            api_func = None
        if api_func is not None:
            from_api = api_func(contrib)
            if from_api is not None:
                contrib.update(from_api)
        contrib['key'] = self._get_key(contrib)
        return contrib

    def _search_api(self, contrib, lookup, api_name, result_format):
        """
        Search an API endpoint for a contributor.

        :param contrib: the full contributor dict
        :param lookup: dict of query params (plus display_name) to send to api.search
        :param api_name: name of the API (see self.api)
        :param result_format: display format of each result, e.g. "{name} ({location})"
        :return: None if not found, dict for matching result if found/selected
        """
        aff = '; '.join(contrib.get('affiliations', []))
        api = self.api[api_name]
        display_name = lookup.pop('display_name')
        try:
            question = (
                f'Do any of these {api_name} results match "{display_name}" ({aff})?'
            )
            click.echo(f'\nSearching {api_name} for "{display_name}"...')
            results = api.search(**lookup).get(u'records', [])
            if len(results) > 0:
                i = multi_choice(
                    question,
                    [result_format.format(**r) for r in results] + ['None of these'],
                    default=len(results),
                )
                if i == len(results):
                    update_dict = self._manual_edit(contrib, api_name)
                else:
                    update_dict = results[i]
            else:
                click.echo(f'No results found for "{display_name}".')
                if click.confirm('Try with a different name?'):
                    new_name = prompt('New name to search: ', default=display_name)
                    if contrib['agent_type'] == 'person':
                        new_name = HumanName(new_name)
                        lookup['family_name'] = new_name.last
                        lookup['given_names'] = new_name.first
                        contrib['family_name'] = new_name.last
                        contrib['given_names'] = new_name.first
                    else:
                        lookup['q'] = new_name
                        lookup['name'] = new_name
                    lookup['display_name'] = self._get_key(contrib)
                    update_dict = self._search_api(
                        contrib, lookup, api_name, result_format
                    )
                else:
                    update_dict = self._manual_edit(contrib, api_name)
        except Exception as e:
            click.echo(f'{api_name} search error for "{display_name}"', err=True)
            click.echo(e, err=True)
            update_dict = self._manual_edit(contrib, api_name)
        update_dict['agent_type'] = contrib['agent_type']
        if 'name' in update_dict or 'family_name' in update_dict:
            new_name = self._get_key(update_dict)
            click.echo(f'Setting name to {new_name}')
        return update_dict

    def _manual_edit(self, contrib, api_name):
        if click.confirm('Enter ID manually?'):
            api = self.api[api_name]
            _id = click.prompt(f'{api_name} ID', show_default=False).strip()
            return api.as_agent_record(api.read(_id))
        if not click.confirm('Edit names manually?'):
            return {}
        if contrib.get('agent_type') == 'person':
            return {
                'given_names': prompt('Given names: ', default=contrib['given_names']),
                'family_name': prompt('Family name: ', default=contrib['family_name']),
            }
        else:
            return {'name': prompt('Name: ', default=contrib['name'])}

    def search_orcid(self, contrib):
        display_name = ' '.join([contrib['given_names'], contrib['family_name']])
        result_format = '{family_name}, {given_names} ({external_id})'
        lookup = {
            'display_name': display_name,
            'family_name': contrib['family_name'],
            'given_names': contrib['given_names'],
        }
        return self._search_api(contrib, lookup, 'ORCID', result_format)

    def search_ror(self, contrib):
        result_format = '{name}, {location} ({external_id})'
        lookup = {'display_name': contrib['name'], 'q': contrib['name']}
        return self._search_api(contrib, lookup, 'ROR', result_format)

    def _get_key(self, contrib_dict):
        if contrib_dict['agent_type'] == 'person':
            return contrib_dict['family_name'] + ', ' + contrib_dict['given_names']
        else:
            return contrib_dict['name']



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.model import DomainObject, meta
from ckan.model.types import make_uuid
from ckanext.attribution.model.agent import agent_table
from ckanext.attribution.model.contribution_activity import contribution_activity_table
from sqlalchemy import Column, ForeignKey, Table, UnicodeText

agent_contribution_activity_table = Table(
    'agent_contribution_activity',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column(
        'agent_id',
        UnicodeText,
        ForeignKey('agent.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=False,
    ),
    Column(
        'contribution_activity_id',
        UnicodeText,
        ForeignKey('contribution_activity.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=False,
        unique=True,
    ),
)


class AgentContributionActivity(DomainObject):
    """
    A link between an agent and the contribution activity performed.
    """

    pass


def check_for_table():
    """
    
    """
    if agent_table.exists() and contribution_activity_table.exists():
        agent_contribution_activity_table.create(checkfirst=True)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from abc import ABCMeta
from ckan.model import DomainObject, Session
from sqlalchemy import Table
from ckan.plugins import toolkit


class BaseQuery(object):
    """
    A base class for easy CRUD (create, read, update, delete) access to attribution
    models.
    """

    __metaclass__ = ABCMeta

    #: :type: The associated database model type.
    m = DomainObject

    #: :sqlalchemy.Table: The associated database table instance
    t = Table()

    @classmethod
    def _columns(cls, **kwargs):
        '''


        :param kwargs:

        '''
        return {c.name: kwargs.get(c.name) for c in cls.t.c if c.name in kwargs}

    @classmethod
    def validate(cls, data_dict):
        """
        Ensure the data_dict provided contains the correct parameters for creating or
        updating a record, and fix issues where possible by deleting extra fields.

        :param data_dict: a complete dictionary of parameters that will be passed to :func:`create`
                          or :func:`update`
        :type data_dict: dict
        :returns: updated data_dict if valid, raises error if not
        """

        def _empty_string_to_null(item):
            if isinstance(item, list):
                return [_empty_string_to_null(i) for i in item]
            if isinstance(item, dict):
                return {k: _empty_string_to_null(v) for k, v in item.items()}
            if item == '':
                return None
            return item

        data_dict = cls._columns(**_empty_string_to_null(data_dict))

        if 'id' in data_dict:
            existing_record = cls.read(data_dict.get('id'))
            if existing_record is None:
                return data_dict
            for c in cls.t.c:
                if c.name not in data_dict:
                    data_dict[c.name] = getattr(existing_record, c.name)

        return data_dict

    @classmethod
    def create(cls, **kwargs):
        """
        Create a new record of type :class:`~m`.
        """
        item_dict = cls._columns(**kwargs)
        new_item = cls.m(**item_dict)
        Session.add(new_item)
        Session.commit()
        return new_item

    @classmethod
    def read(cls, item_id):
        """
        Retrieve a record of type :class:`~m` by its ID.

        :param item_id: the ID of the record.
        :type item_id: str
        """
        retrieved_item = Session.query(cls.m).get(item_id)
        if retrieved_item is None:
            raise toolkit.ObjectNotFound('{0} was not found.'.format(item_id))
        return retrieved_item

    @classmethod
    def exists(cls, item_id):
        """
        Check if a record with the given ID exists.

        :param item_id: the ID of the potential record
        :return: bool
        """
        return Session.query(cls.m).get(item_id) is not None

    @classmethod
    def search(cls, query):
        """
        Retrieve all records matching the search criteria.

        :param query: a sqlalchemy filter query
        """
        return Session.query(cls.m).filter(query).all()

    @classmethod
    def all(cls):
        """
        Return all records.
        """
        return Session.query(cls.m).all()

    @classmethod
    def update(cls, item_id, **kwargs):
        '''


        :param item_id:
        :param kwargs:

        '''
        try:
            del kwargs['id']
        except KeyError:
            pass
        retrieved_item = Session.query(cls.m).filter(cls.m.id == item_id)
        if retrieved_item.count() < 1:
            raise toolkit.ObjectNotFound('{0} was not found.'.format(item_id))
        retrieved_item.update(cls._columns(**kwargs))
        Session.commit()
        return Session.query(cls.m).get(item_id)

    @classmethod
    def delete(cls, item_id):
        '''


        :param item_id:

        '''
        to_delete = Session.query(cls.m).get(item_id)
        if to_delete is not None:
            Session.delete(to_delete)
            Session.commit()



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

import requests
import json


class RorApi(object):
    url = 'https://api.ror.org'

    def read(self, ror_id):
        '''


        :param ror_id:

        '''
        response = requests.get('{0}/organizations/{1}'.format(self.url, ror_id))
        if not response.ok:
            return None
        try:
            record = response.json()
        except json.JSONDecodeError:
            return None
        if record.get('id') is None:
            return None
        return record

    def search(self, q):
        response = requests.get(
            '{0}/organizations'.format(self.url), params={'query': q}
        )
        if not response.ok:
            search_response = {}
        else:
            try:
                search_response = response.json()
            except json.JSONDecodeError:
                search_response = {}
        return {
            'total': search_response.get('number_of_results', 0),
            'records': [
                self.as_agent_record(r) for r in search_response.get('items', [])
            ],
        }

    def as_agent_record(self, ror_record):
        '''


        :param ror_record:

        '''
        if ror_record is None:
            return {}
        address = ror_record.get('addresses')
        if address is not None and len(address) > 0:
            location = address[0].get('city')
        else:
            location = ror_record.get('country', {}).get('country_name')
        return {
            'name': ror_record.get('name', ''),
            'external_id': ror_record.get('id', '').split('/')[-1],
            'external_id_scheme': 'ror',
            'location': location,
            'agent_type': 'org',  # default
        }



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.model import DomainObject, meta
from ckan.model.types import make_uuid
from sqlalchemy import Column, DateTime, Table, UnicodeText, Integer

# this table stores contribution activities
contribution_activity_table = Table(
    'contribution_activity',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column('activity', UnicodeText, nullable=False),
    Column('scheme', UnicodeText, nullable=False),
    Column('level', UnicodeText, nullable=True),
    Column('time', DateTime, nullable=True),
    Column('order', Integer, nullable=True),
)


class ContributionActivity(DomainObject):
    """
    A contribution activity record.
    """

    pass


def check_for_table():
    """
    
    """
    contribution_activity_table.create(checkfirst=True)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

import itertools
import re

from ckan.plugins import toolkit, plugin_loaded
from ckanext.attribution.model.crud import PackageQuery, AgentQuery


def can_edit():
    """
    Check editing permissions for updating agent directly.

    :return:
    """
    try:
        permitted = toolkit.check_access('agent_update', {}, {})
        return permitted
    except toolkit.NotAuthorized:
        return False


def split_caps(string_input):
    return re.sub('(.)(?=[A-Z][^A-Z])', '\\1 ', string_input)


def get_contributions(pkg_id):
    '''
    Template access for the
    :func:`~ckanext.attribution.model.crud.PackageQuery.get_contributions` query method.
    :param pkg_id:
    :return:
    '''
    return PackageQuery.get_contributions(pkg_id)


def get_cited_contributors(pkg_id):
    contributions = PackageQuery.get_contributions(pkg_id)
    by_agent = {
        k: list(v)
        for k, v in itertools.groupby(
            sorted(contributions, key=lambda x: x.agent.id), key=lambda x: x.agent.id
        )
    }

    def _citation_order(entry):
        activities = entry[1]
        citation = [a.order for a in activities if a.activity == '[citation]']
        if citation:
            return max(citation)
        else:
            return -1

    entries = [
        {
            'agent': c[1][0].agent,
            'contributions': [a for a in c[1] if a.activity != '[citation]'],
            'order': _citation_order(c),
        }
        for c in by_agent.items()
    ]

    cited_agents = {
        'cited' if k else 'uncited': sorted(list(v), key=lambda x: x['order'])
        for k, v in itertools.groupby(
            sorted(entries, key=lambda x: x['order']), key=lambda x: x['order'] != -1
        )
    }
    if 'cited' not in cited_agents:
        cited_agents['cited'] = []
    if 'uncited' not in cited_agents:
        cited_agents['uncited'] = []
    return cited_agents


def controlled_list(list_name):
    return toolkit.get_action('attribution_controlled_lists')(
        {}, {'lists': [list_name]}
    )[list_name]


def doi_plugin():
    return plugin_loaded('doi')


def agent_from_user(user_id):
    matches = AgentQuery.search(AgentQuery.m.user_id == user_id)
    if matches:
        return matches[0]
    else:
        return


def user_contributions(user_id):
    agent = agent_from_user(user_id)
    if not agent:
        return []
    pkg_show = toolkit.get_action('package_show')
    by_package = {
        k: list(v)
        for k, v in itertools.groupby(
            sorted(agent.contribution_activities, key=lambda x: x.package.id),
            key=lambda x: x.package.id,
        )
    }
    package_details = []
    for k, v in by_package.items():
        try:
            pkg = pkg_show({}, {'id': k})
        except toolkit.NotAuthorized:
            continue
        activities = [x for x in v if x.activity != '[citation]']
        try:
            citation = next(x.order for x in v if x.activity == '[citation]')
        except StopIteration:
            citation = None
        package_details.append(
            {'dataset': pkg, 'activities': activities, 'citation': citation}
        )
    return package_details



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckanext.attribution.model.contribution_activity import (
    ContributionActivity,
    contribution_activity_table,
)

from ._base import BaseQuery


class ContributionActivityQuery(BaseQuery):
    """
    
    """

    # model and table (subclasses should override)
    m = ContributionActivity
    t = contribution_activity_table



from .agent import AgentQuery
from .agent_affiliation import AgentAffiliationQuery
from .agent_contribution_activity import AgentContributionActivityQuery
from .contribution_activity import ContributionActivityQuery
from .package_contribution_activity import PackageContributionActivityQuery
from .package import PackageQuery



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK


from ckantools.decorators import auth
from ckantools.vars import auth_valid


@auth(anon=True)
def attribution_controlled_lists(context, data_dict):
    return auth_valid


@auth('agent_show')
def agent_external_search(context, data_dict):
    return auth_valid


@auth('agent_show')
def agent_external_read(context, data_dict):
    return auth_valid


@auth(anon=True)
def validate_external_id(context, data_dict):
    return auth_valid



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

import itertools

from ckan.model import Package, package_table, Session
from ckan.plugins import toolkit

from ._base import BaseQuery
from .package_contribution_activity import PackageContributionActivityQuery


class PackageQuery(BaseQuery):
    """
    
    """

    m = Package
    t = package_table

    @classmethod
    def get_contributions(cls, pkg_id):
        pkg = Session.query(Package).get(pkg_id)
        if pkg is None:
            pkg = Session.query(Package).filter(Package.name == pkg_id).first()
        if pkg is None:
            raise toolkit.ObjectNotFound('Package does not exist.')
        else:
            pkg_id = pkg.id

        link_records = PackageContributionActivityQuery.read_package(pkg_id)
        activities = sorted(
            [r.contribution_activity for r in link_records], key=lambda x: x.activity
        )
        return activities

    @classmethod
    def get_agents(cls, pkg_id):
        link_records = PackageContributionActivityQuery.read_package(pkg_id)
        agents = list(
            set(
                sorted(
                    [r.contribution_activity.agent for r in link_records],
                    key=lambda x: x.standardised_name,
                )
            )
        )
        return agents



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

import itertools

from ckanext.attribution.logic.actions.meta import help, schema
from ckanext.attribution.model.crud import (
    AgentAffiliationQuery,
    AgentContributionActivityQuery,
    AgentQuery,
    ContributionActivityQuery,
    PackageContributionActivityQuery,
    PackageQuery,
)
from ckantools.decorators import action
from fuzzywuzzy import fuzz
from sqlalchemy import or_

from ckan.plugins import toolkit


@action(schema.agent_affiliation_show, help.agent_affiliation_show, get=True)
def agent_affiliation_show(original_data_dict):
    item_id = original_data_dict.pop('id')
    return AgentAffiliationQuery.read(item_id).as_dict()


@action(schema.agent_show, help.agent_show, get=True)
def agent_show(original_data_dict):
    item_id = original_data_dict.pop('id')
    return AgentQuery.read(item_id).as_dict()


@action(schema.agent_list, help.agent_list, get=True)
def agent_list(q=None, mode='normal'):
    if q is not None and q != '':
        q_string = '{0}%'.format(q)
        name_cols = [
            AgentQuery.m.name,
            AgentQuery.m.family_name,
            AgentQuery.m.given_names,
        ]
        name_parts = [
            subq
            for c in name_cols
            for subq in [c.ilike('{0}%'.format(p)) for p in q.split(' ')]
        ]
        q_parts = [*name_parts, AgentQuery.m.external_id.ilike(q_string)]
        portal_results = AgentQuery.search(or_(*q_parts))
    else:
        portal_results = AgentQuery.all()
    results = [a.as_dict() for a in portal_results]
    if mode == 'duplicates':
        results = [
            r for r in results if fuzz.token_set_ratio(q, r['display_name']) >= 90
        ]
    return results


@action(
    schema.agent_contribution_activity_show,
    help.agent_contribution_activity_show,
    get=True,
)
def agent_contribution_activity_show(original_data_dict):
    item_id = original_data_dict.pop('id')
    return AgentContributionActivityQuery.read(item_id).as_dict()


@action(schema.contribution_activity_show, help.contribution_activity_show, get=True)
def contribution_activity_show(original_data_dict):
    item_id = original_data_dict.pop('id')
    return ContributionActivityQuery.read(item_id).as_dict()


@action(
    schema.package_contribution_activity_show,
    help.package_contribution_activity_show,
    get=True,
)
def package_contribution_activity_show(original_data_dict):
    item_id = original_data_dict.pop('id')
    return PackageContributionActivityQuery.read(item_id).as_dict()


@action(schema.package_contributions_show, help.package_contributions_show, get=True)
def package_contributions_show(context, original_data_dict, limit=None, offset=0):
    item_id = original_data_dict.pop('id')
    limit = int(limit) if limit is not None else None
    contributions = PackageQuery.get_contributions(item_id)
    by_agent = {
        k: list(v)
        for k, v in itertools.groupby(
            sorted(contributions, key=lambda x: x.agent.id), key=lambda x: x.agent.id
        )
    }
    total = len(by_agent)
    agent_order = [
        (
            {
                'agent': v[0].agent,
                'activities': [a.as_dict() for a in v],
                'affiliations': toolkit.get_action('agent_affiliations')(
                    context, {'agent_id': k, 'package_id': item_id}
                ),
            },
            v[0].agent.package_order(item_id),
        )
        for k, v in by_agent.items()
    ]
    sorted_contributions = [
        c
        for c, o in sorted(
            agent_order,
            key=lambda x: (x[1] if x[1] >= 0 else total, x[0]['agent'].sort_name),
        )
    ]

    page_end = offset + limit if limit is not None else total + 1
    contributions_dict = {
        'contributions': [
            {
                'agent': c['agent'].as_dict(),
                'activities': c['activities'],
                'affiliations': c['affiliations'],
            }
            for c in sorted_contributions[offset:page_end]
        ],
        'all_agents': [c['agent'].id for c in sorted_contributions],
        'total': total,
        'cited_total': len([x for x in agent_order if x[1] >= 0]),
        'offset': offset,
        'page_size': limit or total,
    }
    return contributions_dict


@action(schema.agent_affiliations, help.agent_affiliations, get=True)
def agent_affiliations(agent_id, package_id):
    agent = AgentQuery.read(agent_id)
    affiliations = agent.affiliations
    if package_id is not None:
        affiliations = [
            a
            for a in affiliations
            if a['affiliation'].package_id is None
            or a['affiliation'].package_id == package_id
        ]

    def _transform(a):
        detail = a.get('affiliation').as_dict()
        try:
            del detail['agent_a_id']
        except KeyError:
            pass
        try:
            del detail['agent_b_id']
        except KeyError:
            pass
        detail['other_agent'] = a.get('agent').as_dict()
        return detail

    return [_transform(a) for a in affiliations]



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckanext.attribution.logic.actions.meta import help, schema
from ckanext.attribution.model.crud import (
    AgentAffiliationQuery,
    AgentContributionActivityQuery,
    AgentQuery,
    ContributionActivityQuery,
    PackageContributionActivityQuery,
)
from ckantools.decorators import action


@action(schema.agent_affiliation_delete, help.agent_affiliation_delete)
def agent_affiliation_delete(original_data_dict):
    item_id = original_data_dict.pop('id')
    return AgentAffiliationQuery.delete(item_id)


@action(schema.agent_delete, help.agent_delete)
def agent_delete(original_data_dict):
    item_id = original_data_dict.pop('id')
    return AgentQuery.delete(item_id)


@action(
    schema.agent_contribution_activity_delete, help.agent_contribution_activity_delete
)
def agent_contribution_activity_delete(original_data_dict):
    item_id = original_data_dict.pop('id')
    return AgentContributionActivityQuery.delete(item_id)


@action(schema.contribution_activity_delete, help.contribution_activity_delete)
def contribution_activity_delete(original_data_dict):
    item_id = original_data_dict.pop('id')
    return ContributionActivityQuery.delete(item_id)


@action(
    schema.package_contribution_activity_delete,
    help.package_contribution_activity_delete,
)
def package_contribution_activity_delete(original_data_dict):
    item_id = original_data_dict.pop('id')
    return PackageContributionActivityQuery.delete(item_id)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckantools.decorators import auth
from ckantools.vars import auth_valid


@auth()
def agent_affiliation_create(context, data_dict):
    """
    Allow for logged-in users.
    """
    return auth_valid


@auth()
def agent_create(context, data_dict):
    """
    Allow for logged-in users.
    """
    return auth_valid


@auth()
def contribution_activity_create(context, data_dict):
    """
    Allow for logged-in users.
    """
    return auth_valid



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK
import re

from ckan.plugins import toolkit
from ckanext.attribution.lib.orcid_api import OrcidApi
from ckanext.attribution.lib.ror_api import RorApi
from ckanext.attribution.model.crud import AgentQuery
from ckantools.decorators import action
from ckanext.attribution.logic.actions.meta import schema, help


@action(
    schema.attribution_controlled_lists, help.attribution_controlled_lists, get=True
)
def attribution_controlled_lists(context, lists=None):
    all_lists = {
        'agent_types': {
            'person': {'fa_icon': 'fas fa-user', 'default_scheme': 'orcid'},
            'org': {'fa_icon': 'fas fa-building', 'default_scheme': 'ror'},
            'other': {'fa_icon': 'fas fa-asterisk', 'default_scheme': None},
        },
        'contribution_activity_types': {
            'credit': [
                {'name': 'Conceptualization'},
                {'name': 'Data curation'},
                {'name': 'Formal analysis'},
                {'name': 'Funding acquisition'},
                {'name': 'Investigation'},
                {'name': 'Methodology'},
                {'name': 'Project administration'},
                {'name': 'Resources'},
                {'name': 'Software'},
                {'name': 'Supervision'},
                {'name': 'Validation'},
                {'name': 'Visualization'},
                {'name': 'Writing – original draft'},
                {'name': 'Writing – review & editing'},
            ],
            'datacite': [
                {'name': 'ContactPerson'},
                {'name': 'DataCollector'},
                {'name': 'DataCurator'},
                {'name': 'DataManager'},
                {'name': 'Distributor'},
                {'name': 'Editor'},
                {'name': 'HostingInstitution'},
                {'name': 'Other'},
                {'name': 'Producer'},
                {'name': 'ProjectLeader'},
                {'name': 'ProjectManager'},
                {'name': 'ProjectMember'},
                {'name': 'RegistrationAgency'},
                {'name': 'RegistrationAuthority'},
                {'name': 'RelatedPerson'},
                {'name': 'ResearchGroup'},
                {'name': 'RightsHolder'},
                {'name': 'Researcher'},
                {'name': 'Sponsor'},
                {'name': 'Supervisor'},
                {'name': 'WorkPackageLeader'},
            ],
            'internal': [{'name': 'Unspecified'}],
        },
        'contribution_activity_levels': ['Lead', 'Supporting', 'Equal'],
        'agent_external_id_schemes': {
            'orcid': {
                'url': 'https://orcid.org/{0}',
                'scheme_uri': 'https://orcid.org',
                'label': 'ORCID',
                'fa_icon': 'fab fa-orcid',
                'rgx': r'(?:\d{4}-){3}\d{3}[\dX]',
            },
            'ror': {
                'url': 'https://ror.org/{0}',
                'scheme_uri': 'https://ror.org',
                'label': 'ROR',
                'fa_icon': 'fas fa-university',
                'rgx': r'0[0-9a-hjkmnp-z]{6}\d{2}',
            },
        },
    }
    if lists is not None and isinstance(lists, str):
        lists = [l.strip() for l in lists.split(',')]
    if lists is not None and isinstance(lists, list):
        lists = [l for l in lists if l in all_lists]
        return {k: v for k, v in all_lists.items() if k in lists}
    else:
        return all_lists


@action(schema.agent_external_search, help.agent_external_search, get=True)
def agent_external_search(q, sources):
    if sources is not None and isinstance(sources, str):
        sources = [sources.lower()]
    elif sources is not None and isinstance(sources, list):
        sources = [s.lower() for s in sources]
    elif sources is not None:
        sources = None
    results = {}
    if q is None or q == '':
        return results

    # ORCID
    if sources is None or 'orcid' in sources:
        orcid_remaining = 0
        orcidapi = OrcidApi()
        orcid_search = orcidapi.search(q=q)
        n = orcid_search.get(u'total', 0)
        orcid_records = orcid_search.get('records')
        orcid_records = [
            r
            for r in orcid_records
            if AgentQuery.read_external_id(r['external_id']) is None
        ]
        if n > len(orcid_records):
            orcid_remaining = n - len(orcid_records)
        results['orcid'] = {'records': orcid_records, 'remaining': orcid_remaining}

    # ROR
    if sources is None or 'ror' in sources:
        ror_remaining = 0
        rorapi = RorApi()
        ror_search = rorapi.search(q=q)
        n = ror_search.get(u'total', 0)
        ror_records = ror_search.get('records')
        ror_records = [
            r
            for r in ror_records
            if AgentQuery.read_external_id(r['external_id']) is None
        ]
        if n > len(ror_records):
            ror_remaining = n - len(ror_records)
        results['ror'] = {'records': ror_records, 'remaining': ror_remaining}

    return results


@action(schema.agent_external_read, help.agent_external_read, get=True)
def agent_external_read(agent_id, external_id, external_id_scheme, diff=False):
    if agent_id is None and (external_id is None or external_id_scheme is None):
        raise toolkit.Invalid(
            'Either record ID or external ID + scheme must be provided.'
        )
    if agent_id is not None:
        updated_dict = AgentQuery.read_from_external_api(agent_id)
        updated_dict['id'] = agent_id
        del updated_dict[
            'user_id'
        ]  # this is internal only so it's always going to be None
        if diff:
            item_dict = AgentQuery.read(agent_id).as_dict()
            for k, v in item_dict.items():
                if k in updated_dict and updated_dict.get(k) == v:
                    del updated_dict[k]
        return updated_dict
    else:
        apis = {'orcid': OrcidApi, 'ror': RorApi}
        api = apis[external_id_scheme.lower()]()
        return api.as_agent_record(api.read(external_id))


@action(schema.validate_external_id, help.validate_external_id)
def validate_external_id(context, external_id, external_id_scheme):
    # extract via regex first
    controlled_lists = toolkit.get_action('attribution_controlled_lists')(context, {})
    scheme = controlled_lists['agent_external_id_schemes'][external_id_scheme.lower()]
    rgx = re.compile(scheme['rgx'])
    matches = rgx.findall(external_id)
    if not matches:
        return {'id': None, 'error': 'No valid ID string found.'}
    api = {'orcid': OrcidApi, 'ror': RorApi}[external_id_scheme.lower()]()
    for match in matches:
        try:
            record = api.read(match)
            if record:
                return {'id': match, 'error': None}
        except Exception as e:
            continue
    return {'id': None, 'error': '{0} ID could not be found'.format(external_id_scheme)}



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit
from ckanext.attribution.logic.actions.helpers import (
    parse_contributors,
    get_author_string,
)
from ckanext.attribution.model.crud import (
    AgentQuery,
    ContributionActivityQuery,
    AgentAffiliationQuery,
    AgentContributionActivityQuery,
)
from ckantools.decorators import action, basic_action
from ckanext.attribution.logic.actions.meta import help, schema


@action(schema.agent_affiliation_update, help.agent_affiliation_update)
def agent_affiliation_update(context, original_data_dict, agent_a_id, agent_b_id):
    item_id = original_data_dict.pop('id')
    # check agents exist if updating
    for agent_id in [agent_a_id, agent_b_id]:
        if agent_id is None:
            continue
        try:
            toolkit.get_action('agent_show')(context, {'id': agent_id})
        except toolkit.ObjectNotFound:
            raise toolkit.ValidationError(
                'Agent ({0}) does not exist.'.format(agent_id)
            )
    affiliation = AgentAffiliationQuery.update(item_id, **original_data_dict)
    if affiliation is None:
        raise toolkit.ValidationError(
            'Unable to update affiliation. Check the fields are valid.'
        )
    return affiliation.as_dict()


@action(schema.agent_update, help.agent_update)
def agent_update(original_data_dict, agent_type=None):
    item_id = original_data_dict.get('id')
    current_record = AgentQuery.read(item_id)
    old_citation_name = current_record.citation_name
    if agent_type is None:
        agent_type = current_record.agent_type
    original_data_dict['agent_type'] = agent_type
    data_dict = AgentQuery.validate(original_data_dict)
    new_agent = AgentQuery.update(item_id, **original_data_dict)
    if new_agent.citation_name != old_citation_name:
        # if the name has been updated, the author strings need to be updated everywhere else too
        agent_id_column = AgentContributionActivityQuery.m.agent_id
        contrib_activities = AgentContributionActivityQuery.search(
            agent_id_column == item_id
        )
        packages = list(
            set([c.contribution_activity.package.id for c in contrib_activities])
        )
        for p in packages:
            author_string = get_author_string(package_id=p)
            toolkit.get_action('package_revise')(
                {}, {'match': {'id': p}, 'update': {'author': author_string}}
            )
    if new_agent is None:
        raise toolkit.ValidationError(
            'Unable to update agent. Check the fields are valid.'
        )
    return new_agent.as_dict()


@action(schema.agent_external_update, help.agent_external_update)
def agent_external_update(original_data_dict):
    item_id = original_data_dict.pop('id')
    updated_dict = AgentQuery.read_from_external_api(item_id)
    updated_agent = AgentQuery.update(item_id, **updated_dict)
    if updated_agent is None:
        raise toolkit.ValidationError(
            'Unable to update agent. Check the fields are valid.'
        )
    return updated_agent.as_dict()


@action(schema.contribution_activity_update, help.contribution_activity_update)
def contribution_activity_update(original_data_dict):
    item_id = original_data_dict.get('id')
    data_dict = ContributionActivityQuery.validate(original_data_dict)
    new_activity = ContributionActivityQuery.update(item_id, **data_dict)
    return new_activity.as_dict()


@toolkit.chained_action
@basic_action
def package_update(next_func, context, data_dict):
    parse_contributors(context, data_dict)
    data_dict['author'] = get_author_string(package_id=data_dict['id'])
    return next_func(context, data_dict)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.model import DomainObject, meta
from ckan.model.types import make_uuid
from ckanext.attribution.model.agent import agent_table
from sqlalchemy import Column, Date, ForeignKey, Table, UnicodeText

agent_affiliation_table = Table(
    'agent_affiliation',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column(
        'agent_a_id',
        UnicodeText,
        ForeignKey('agent.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=False,
    ),
    Column(
        'agent_b_id',
        UnicodeText,
        ForeignKey('agent.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=False,
    ),
    Column('affiliation_type', UnicodeText, nullable=True),
    Column('description', UnicodeText, nullable=True),
    Column('start_date', Date, nullable=True),
    Column('end_date', Date, nullable=True),
    Column(
        'package_id',
        UnicodeText,
        ForeignKey('package.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=True,
    ),
)


class AgentAffiliation(DomainObject):
    """
    An affiliation between two agents (e.g. agent a is a researcher, agent b is a
    university).
    """

    @property
    def agents(self):
        return [self._agent_a, self._agent_b]

    def other_agent(self, agent_id):
        assert agent_id in [self.agent_a_id, self.agent_b_id]
        return self._agent_a if agent_id == self.agent_b_id else self._agent_b


def check_for_table():
    """
    
    """
    if agent_table.exists():
        agent_affiliation_table.create(checkfirst=True)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckanext.attribution.model.agent_contribution_activity import (
    AgentContributionActivity,
    agent_contribution_activity_table,
)

from ._base import BaseQuery


class AgentContributionActivityQuery(BaseQuery):
    """
    
    """

    # model and table (subclasses should override)
    m = AgentContributionActivity
    t = agent_contribution_activity_table

    @classmethod
    def read_agent_package(cls, agent_id, package_id):
        return [
            r.contribution_activity
            for r in cls.search(cls.m.agent_id == agent_id)
            if r.contribution_activity.package.id == package_id
        ]



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckantools.decorators import auth
from ckantools.vars import auth_valid
from ckan.plugins import toolkit


@auth(anon=True)
def agent_affiliation_show(context, data_dict):
    """
    Allow for everyone.
    """
    return auth_valid


@auth(anon=True)
def agent_show(context, data_dict):
    """
    Allow for everyone.
    """
    return auth_valid


@auth(proxy='agent_show', anon=True)
def agent_list(context, data_dict):
    return auth_valid


@auth(anon=True)
def agent_contribution_activity_show(context, data_dict):
    """
    Allow for everyone.
    """
    return auth_valid


@auth(anon=True)
def contribution_activity_show(context, data_dict):
    """
    Allow for everyone.
    """
    return auth_valid


@auth(anon=True)
def package_contribution_activity_show(context, data_dict):
    """
    Allow for everyone.
    """
    return auth_valid


@auth(anon=True)
def package_contributions_show(context, data_dict):
    """
    Allow for everyone.
    """
    return auth_valid


@auth(proxy=['agent_show', 'agent_affiliation_show'], anon=True)
def agent_affiliations(context, data_dict):
    # TODO: remove anon=True once ckantools#9 is fixed
    return auth_valid



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

# CREATE ===========================================================================================

agent_affiliation_create = '''
Create an :class:`~ckanext.attribution.model.agent_affiliation.AgentAffiliation` link record
between two :class:`~ckanext.attribution.model.agent.Agent` records, e.g. to show institutional
affiliation for an author.

:param agent_a_id: ID of the first agent
:type agent_a_id: str
:param agent_b_id: ID of the second agent
:type agent_b_id: str
:param affiliation_type: type of affiliation/relationship between the two agents
:type affiliation_type: str, optional
:param description: description of affiliation/relationship
:type description: str, optional
:param start_date: when the affiliation started (e.g. when employment began)
:type start_date: datetime.date, optional
:param end_date: when the affiliation ended (e.g. when a researcher left an institution)
:type end_date: datetime.date, optional
:returns: New agent affiliation record.
:rtype: dict
'''

agent_create = '''
Action for creating an :class:`~ckanext.attribution.model.agent.Agent` record. Different fields are
required by different agent types.

:param agent_type: broad type of agent; usually 'person' or 'org'
:type agent_type: str
:param family_name: family name of an person [person only, required]
:type family_name: str, optional
:param given_names: given name(s) or initials of an person [person only, required]
:type given_names: str, optional
:param given_names_first: whether given names should be displayed before the family name
                          (default True) [person only, optional]
:type given_names_first: bool, optional
:param user_id: the ID for a registered user of this CKAN instance associated with this agent
                [person only, optional]
:type user_id: str, optional
:param name: name of an organisation [org only, required]
:type name: str, optional
:returns: New agent record.
:rtype: dict
'''

contribution_activity_create = '''
Creates a :class:`~ckanext.attribution.model.contribution_activity.ContributionActivity` record,
linked to a package and an agent via package_contribution_activity and agent_contribution_activity
records (respectively). These link records are also created as part of this action, as the activity
should not exist without the package or agent.

:param package_id: the ID for the package this activity is associated with
:type package_id: str
:param agent_id: the ID for the agent this activity is associated with
:type agent_id: str
:param activity: short (one/two words) description for the activity
:type activity: str
:param scheme: name of the role/activity taxonomy, e.g. credit or datacite
:type scheme: str
:param level: lead, equal, or supporting
:type level: str, optional
:param time: time activity took place
:type time: datetime.datetime, optional
:returns: New contribution activity record.
:rtype: dict
'''

# DELETE ===========================================================================================

agent_affiliation_delete = '''
Delete an :class:`~ckanext.attribution.model.agent_affiliation.AgentAffiliation` record by ID.

:param id: ID of the affiliation record
:type id: str
:returns: The affiliation record.
:rtype: dict
'''

agent_delete = '''
Delete an :class:`~ckanext.attribution.model.agent.Agent` record by ID.

:param id: ID of the agent record
:type id: str
:returns: The agent record.
:rtype: dict
'''

agent_contribution_activity_delete = '''
Delete an :class:`~ckanext.attribution.model.agent_contribution_activity.AgentContributionActivity`
record by ID.

:param id: ID of the agent contribution activity record
:type id: str
:returns: The agent contribution activity record.
:rtype: dict
'''

contribution_activity_delete = '''
Delete a :class:`~ckanext.attribution.model.contribution_activity.ContributionActivity` record by
ID.

:param id: ID of the contribution activity record
:type id: str
:returns: The contribution activity record.
:rtype: dict
'''

package_contribution_activity_delete = '''
Delete a
:class:`~ckanext.attribution.model.package_contribution_activity.PackageContributionActivity` record
by ID.

:param id: ID of the package contribution activity record
:type id: str
:returns: The package contribution activity record.
:rtype: dict
'''

# EXTRA ============================================================================================

attribution_controlled_lists = '''
Return one or more lists or dicts of defined values, e.g. agent types or contribution activity
types. Details dicts can include various pieces of arbitrary information (e.g. names,
translations, or icon definitions for templates) as long as the initial structure is retained.

:param lists: names of the lists to be returned
:type lists: list, optional
:return: dict of all requested lists (or all lists if unspecified)
:rtype: dict
'''

agent_external_search = '''
Search external sources for agent data. Ignores records that already exist in the database.

:param q: searches all fields (names, ids, etc)
:type q: str
:param sources: a list of sources to search (default None searches all)
:type sources: list, optional
:returns: A list of potential matches.
:rtype: list
'''

agent_external_read = '''
Read data from an external source like ORCID or ROR.

:param agent_id: ID of the record to read
:type agent_id: str, optional
:param external_id: ID from external service
:type external_id: str, optional
:param external_id_scheme: external scheme, e.g. orcid
:type external_id_scheme: str, optional
:param diff: only show values that differ from the record's current values; only valid if record
             already exists (default False)
:type diff: bool, optional
:returns: The details pulled from the external source, formatted as a record dict
:rtype: dict
'''

validate_external_id = '''
Validate/format an external ID.

:param external_id: ID from external service
:type external_id: str
:param external_id_scheme: external scheme, e.g. orcid
:type external_id_scheme: str
:return:
'''

# SHOW =============================================================================================

agent_affiliation_show = '''
Retrieve an :class:`~ckanext.attribution.model.agent_affiliation.AgentAffiliation` record by ID.

:param id: ID of the affiliation record
:type id: str
:returns: The affiliation record.
:rtype: dict
'''

agent_show = '''
Retrieve an :class:`~ckanext.attribution.model.agent.Agent` record by ID.

:param id: ID of the agent record
:type id: str
:returns: The agent record.
:rtype: dict
'''

agent_list = '''
Search for :class:`~ckanext.attribution.model.agent.Agent` records.

:param q: name or external id (ORCID/ROR ID) of the agent record
:type q: str, optional
:param mode: 'normal' or 'duplicates'
:type mode: str, optional
:returns: A list of potential matches.
:rtype: list
'''

agent_contribution_activity_show = '''
Retrieve an
:class:`~ckanext.attribution.model.agent_contribution_activity.AgentContributionActivity` record
by ID.

:param id: ID of the agent contribution activity record
:type id: str
:returns: The agent contribution activity record.
:rtype: dict
'''

contribution_activity_show = '''
Retrieve a :class:`~ckanext.attribution.model.contribution_activity.ContributionActivity` record by
ID.

:param id: ID of the contribution activity record
:type id: str
:returns: The contribution activity record.
:rtype: dict
'''

package_contribution_activity_show = '''
Retrieve a
:class:`~ckanext.attribution.model.package_contribution_activity.PackageContributionActivity`
record by ID.

:param id: ID of the package contribution activity record
:type id: str
:returns: The package contribution activity record.
:rtype: dict
'''

package_contributions_show = '''
Show associated agents and their contributions for a given package. Agents are returned in citation
then agent id order.

:param id: ID of the package record
:type id: str
:param limit: limit the number of records returned
:type limit: int
:param offset: skip n agents
:type offset: int
:returns: The package contribution activity record.
:rtype: dict
'''


agent_affiliations = '''
Show affiliated agents, either all or for a given package. Including a package ID will still return
'global' affiliations, e.g. those with no specific package associated.

:param agent_id: ID of the agent
:type agent_id: str
:param package_id: ID of the package
:type package_id: str, optional
:returns: The package contribution activity record.
:rtype: dict
'''

# UPDATE ===========================================================================================

agent_affiliation_update = '''
Update an :class:`~ckanext.attribution.model.agent_affiliation.AgentAffiliation` link record.

:param id: ID of the record to update
:type id: str
:param agent_a_id: ID of the first agent
:type agent_a_id: str, optional
:param agent_b_id: ID of the second agent
:type agent_b_id: str, optional
:param affiliation_type: type of affiliation/relationship between the two agents
:type affiliation_type: str, optional
:param description: description of affiliation/relationship
:type description: str, optional
:param start_date: when the affiliation started (e.g. when employment began)
:type start_date: datetime.date, optional
:param end_date: when the affiliation ended (e.g. when a researcher left an institution)
:type end_date: datetime.date, optional
:param context:
:param data_dict:
:returns: The updated agent affiliation record.
:rtype: dict
'''

agent_update = '''
Action for updating an :class:`~ckanext.attribution.model.agent.Agent` record. Different fields are
required by different agent types.

:param id: ID of the record to update
:type id: str
:param agent_type: broad type of agent; usually 'person' or 'org'
:type agent_type: str, optional
:param family_name: family name of an person [person only]
:type family_name: str, optional
:param given_names: given name(s) or initials of an person [person only]
:type given_names: str, optional
:param given_names_first: whether given names should be displayed before the family name
                          (default True) [person only]
:type given_names_first: bool, optional
:param user_id: the ID for a registered user of this CKAN instance associated with this agent
                [person only]
:type user_id: str, optional
:param name: name of an organisation [org only]
:type name: str, optional
:param context:
:param data_dict:
:returns: The updated agent record.
:rtype: dict
'''

agent_external_update = '''
Action for updating an :class:`~ckanext.attribution.model.agent.Agent` record by pulling information
from an external source like ORCID or ROR.

:param id: ID of the record to update
:type id: str
:returns: The updated agent record.
:rtype: dict
'''

contribution_activity_update = '''
Updates a :class:`~ckanext.attribution.model.contribution_activity.ContributionActivity` record,
linked to a package and an agent via package_contribution_activity and agent_contribution_activity
records (respectively). These link records are also updated as part of this action, as the activity
should not exist without the package or agent.

:param id: ID of the record to update
:type id: str
:param activity: short (one/two words) description for the activity
:type activity: str, optional
:param scheme: name of the role/activity taxonomy, e.g. credit or datacite
:type scheme: str, optional
:param level: lead, equal, or supporting
:type level: str, optional
:param time: time activity took place
:type time: datetime.datetime, optional
:param context:
:param data_dict:
:returns: New contribution activity record.
:rtype: dict
'''



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

import orcid
import requests
from ckan.plugins import toolkit
from werkzeug.utils import cached_property


class OrcidApi(object):
    """
    
    """

    def __init__(self):
        self.key = toolkit.config.get('ckanext.attribution.orcid_key')
        self.secret = toolkit.config.get('ckanext.attribution.orcid_secret')
        self._debug = toolkit.config.get('ckanext.attribution.debug', 'True') == 'True'

    @cached_property
    def conn(self):
        """
        
        """
        if self.key is None or self.secret is None:
            raise Exception(toolkit._('ORCID API credentials not supplied.'))
        return orcid.PublicAPI(self.key, self.secret, sandbox=self._debug)

    @cached_property
    def read_token(self):
        """
        
        """
        if self.key is None or self.secret is None:
            raise Exception(toolkit._('ORCID API credentials not supplied.'))
        url = (
            'https://sandbox.orcid.org/oauth/token'
            if self._debug
            else 'https://orcid.org/oauth/token'
        )
        r = requests.post(
            url,
            data={
                'client_id': self.key,
                'client_secret': self.secret,
                'grant_type': 'client_credentials',
                'scope': '/read-public',
            },
            headers={'Accept': 'application/json'},
        )
        if r.ok:
            return r.json()['access_token']
        else:
            return None

    def search(self, orcid_q=None, q=None, family_name=None, given_names=None):
        '''

        :param orcid_q:  (Default value = None)
        :param q:  (Default value = None)

        '''
        query = []
        if orcid_q is not None and orcid_q != '':
            query.append('orcid:' + orcid_q)
        if q is not None and q != '':
            query.append('text:' + q)
        if family_name is not None and family_name != '':
            query.append('family-name:' + family_name)
        if given_names is not None and given_names != '':
            query.append('given-names:' + given_names)
        query = '+AND+'.join(query)
        search_response = self.conn.search(query, access_token=self.read_token, rows=10)
        records = []
        loaded_ids = []
        for r in search_response.get('result', []):
            _id = r.get(u'orcid-identifier', {}).get(u'path', None)
            if _id is not None and _id not in loaded_ids:
                try:
                    orcid_record = self.as_agent_record(self.read(_id))
                    records.append(orcid_record)
                    loaded_ids.append(_id)
                except AttributeError as e:
                    # probably a malformed orcid record
                    continue
        result = {'total': search_response.get('num-found', 0), 'records': records}
        return result

    def read(self, orcid_id):
        '''


        :param orcid_id:

        '''
        record = self.conn.read_record_public(orcid_id, 'record', self.read_token)
        return record

    def as_agent_record(self, orcid_record):
        '''


        :param orcid_record:

        '''
        names = orcid_record.get('person', {}).get('name', {})
        return {
            'family_name': names.get('family-name', {}).get('value', ''),
            'given_names': names.get('given-names', {}).get('value', ''),
            'external_id': orcid_record.get('orcid-identifier', {}).get('path', ''),
            'external_id_scheme': 'orcid',
            'agent_type': 'person',  # default
        }



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

import itertools
import re

import click
from ckanext.attribution.lib.orcid_api import OrcidApi
from ckanext.attribution.lib.ror_api import RorApi
from fuzzywuzzy import process
from prompt_toolkit import prompt
from unidecode import unidecode

from .common import multi_choice


class Combiner(object):
    """
    Combines names extracted by a Parser.
    """

    def __init__(self, parser):
        self.contributors = parser.contributors
        self.affiliations = parser.affiliations

    def separate(self, group):
        """
        Ensure that the automated grouping is correct.

        :param group: a list of ParsedSegment instances that are probably the same contributor
        :return: a list of lists of ParsedSegments
        """
        all_names = sorted(
            list(set([str(x.name) for x in group])), key=lambda x: -len(x)
        )
        if len(all_names) > 1:
            same = click.confirm(
                'Are these all the same contributor?\n\t{0}\n'.format(
                    '\n\t'.join(all_names)
                ),
                default=True,
            )
            if not same:
                subgroups = {}
                for n in all_names:
                    v = [x for x in group if x.name == n]
                    if len(subgroups) == 0:
                        subgroups[n] = v
                        continue
                    matches = [m[0] for m in process.extract(n, list(subgroups.keys()))]
                    ix = multi_choice(
                        'Is "{0}" the same as any of these contributors?'.format(n),
                        matches + ['None of these'],
                        default=len(matches),
                    )
                    k = matches[ix] if ix < len(matches) else n
                    subgroups[k] = subgroups.get(k, []) + v
                return list(subgroups.values())
        return [group]

    def combine(self, group, agent_type, name_func=None):
        all_names = [x.name for x in group]
        _contrib_dicts = sorted(
            [(ct, pkgs) for c in group for ct, pkgs in c.packages.items()],
            key=lambda x: x[0],
        )
        _grouped_contribs = itertools.groupby(_contrib_dicts, key=lambda x: x[0])
        contrib = {
            'agent_type': agent_type,
            'all_names': [str(n) for n in all_names],
            'affiliations': list(set([a for x in group for a in x.affiliations])),
            'packages': {
                contrib_type: list(set([pkgid for ct, pkgid in v]))
                for contrib_type, v in _grouped_contribs
            },
        }
        if name_func is None:
            longest_name = sorted(list(set(all_names)), key=lambda x: -len(x))[
                0
            ].strip()
            name = {'name': longest_name}
        else:
            name = name_func(all_names)
        contrib.update(name)
        contrib['key'] = self._get_key(contrib)
        return contrib

    def combine_person_names(self, names):
        """
        Uses a list of HumanNames to determine the longest possible name for a person.

        :return: a dict of family_name, given_names (includes middle names), and key (i.e. a sort/display name)
        """

        def _filter_diacritics(name_list):
            filtered = [n for n in name_list if unidecode(n) != n]
            if len(filtered) > 0:
                return filtered
            else:
                return name_list

        given = []
        family = []
        for n in names:
            given.append(' '.join([n.first, n.middle]))
            family.append(n.last)
        given = list(set(given))
        family = list(set(family))

        # use longest family name
        family_name = sorted(_filter_diacritics(family), key=lambda x: -len(x))[
            0
        ].strip()
        # given names are more complicated
        # remove empty strings and split into parts
        given = [re.split(r'\s+', m) for m in list(set(given)) if m != '']
        given_parts = {}
        for m in given:
            for i, x in enumerate(m):
                given_parts[i] = given_parts.get(i, []) + [x]
        given_names = ' '.join(
            [
                sorted(_filter_diacritics(p), key=lambda x: -len(x))[0]
                for p in given_parts.values()
            ]
        ).strip()
        combined = {'family_name': family_name, 'given_names': given_names}
        return combined

    def update_affiliations(self, contributor):
        """
        Update the self.affiliations dict to ensure the names are consistent.

        :param contributor: contributor dict
        """
        no_affiliations = len(contributor.get('affiliations', [])) == 0
        is_not_affiliation = len(contributor['packages'].get('affiliations', [])) == 0
        if no_affiliations and is_not_affiliation:
            return
        all_packages = [
            pkg_id for x in contributor['packages'].values() for pkg_id in x
        ]
        for pkg in all_packages:
            items = self.affiliations.get(pkg[0])
            if items is None:
                continue
            updated_items = []
            for name, affiliation in items:
                if name in contributor['all_names']:
                    updated_items.append((contributor['key'], affiliation))
                elif affiliation in contributor['all_names']:
                    updated_items.append((name, contributor['key']))
                else:
                    updated_items.append((name, affiliation))
            self.affiliations[pkg[0]] = updated_items

    def run(self):
        """
        Run the combiner over the whole parser list, including separating groups,
        combining names, searching APIs, and updating the affiliations dict.

        :return: a list of contributors
        """
        combined = []

        for g in [
            grp
            for family_name, initials_list in self.contributors['person'].items()
            for initial, grp in initials_list.items()
        ]:
            for person in self.separate(g):
                c = self.combine(person, 'person', self.combine_person_names)
                if c is not None:
                    combined.append(c)
                    self.update_affiliations(c)
        for abbr, g in self.contributors['org'].items():
            for org in self.separate(g):
                c = self.combine(org, 'org', None)
                if c is not None:
                    combined.append(c)
                    self.update_affiliations(c)
        for abbr, g in self.contributors['other'].items():
            for o in self.separate(g):
                c = self.combine(o, 'other')
                if c is not None:
                    combined.append(c)
                    self.update_affiliations(c)
        return combined

    def _get_key(self, contrib_dict):
        if contrib_dict['agent_type'] == 'person':
            return contrib_dict['family_name'] + ', ' + contrib_dict['given_names']
        else:
            return contrib_dict['name']



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckan.model import DomainObject, meta, user_table
from ckan.model.types import make_uuid
from ckan.plugins import toolkit
from sqlalchemy import Boolean, Column, ForeignKey, Table, UnicodeText

# this table stores agent
agent_table = Table(
    'agent',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column('agent_type', UnicodeText, nullable=False),
    Column('family_name', UnicodeText, nullable=True),
    Column('given_names', UnicodeText, nullable=True),
    Column('given_names_first', Boolean, nullable=True, default=True),
    Column('name', UnicodeText, nullable=True),
    Column('location', UnicodeText, nullable=True),
    Column('external_id', UnicodeText, nullable=True, unique=True),
    Column('external_id_scheme', UnicodeText, nullable=True),
    Column(
        'user_id',
        UnicodeText,
        ForeignKey('user.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=True,
    ),
)


class Agent(DomainObject):
    """
    An agent (e.g. a researcher or institution) that contributes to a package.
    """

    @property
    def display_name(self):
        if self.agent_type == 'person':
            return ' '.join(
                [self.given_names, self.family_name]
                if self.given_names_first
                else [self.family_name, self.given_names]
            )
        elif self.agent_type == 'org':
            name = self.name
            if self.location is not None:
                name += ' ({0})'.format(self.location)
            return name
        else:
            return self.name

    @property
    def standardised_name(self):
        """
        Family name first for persons.

        :return:
        """
        if self.agent_type == 'person':
            return ', '.join([self.family_name, self.given_names])
        else:
            return self.name

    @property
    def citation_name(self):
        return self.display_name if self.agent_type == 'person' else self.name

    @property
    def external_id_url(self):
        if self.external_id is None:
            return
        external_scheme_dict = toolkit.get_action('attribution_controlled_lists')(
            {}, {'lists': ['agent_external_id_schemes']}
        )['agent_external_id_schemes']
        return external_scheme_dict[self.external_id_scheme]['url'].format(
            self.external_id
        )

    @property
    def affiliations(self):
        return [
            {'agent': a.other_agent(self.id), 'affiliation': a}
            for a in self._affiliations
        ]

    def package_affiliations(self, pkg_id):
        return [
            a
            for a in self.affiliations
            if a['affiliation'].package_id is None
            or a['affiliation'].package_id == pkg_id
        ]

    def package_order(self, pkg_id):
        try:
            citation = next(
                c
                for c in self.contribution_activities
                if c.package
                and (c.package.id == pkg_id or c.package.name == pkg_id)
                and c.activity == '[citation]'
            )
        except StopIteration:
            return -1
        return citation.order

    @property
    def sort_name(self):
        return self.family_name if self.agent_type == 'person' else self.name

    def as_dict(self):
        agent_dict = super(Agent, self).as_dict()
        agent_dict['display_name'] = self.display_name
        agent_dict['standardised_name'] = self.standardised_name
        agent_dict['external_id_url'] = self.external_id_url
        return agent_dict


def check_for_table():
    """
    Checks to see if the user_table exists and creates it if it doesn't.
    """
    if user_table.exists():
        agent_table.create(checkfirst=True)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-attribution
# Created by the Natural History Museum in London, UK

from ckanext.attribution.model.package_contribution_activity import (
    PackageContributionActivity,
    package_contribution_activity_table,
)
from ckan.model import Session
from ._base import BaseQuery


class PackageContributionActivityQuery(BaseQuery):
    """
    
    """

    # model and table (subclasses should override)
    m = PackageContributionActivity
    t = package_contribution_activity_table

    @classmethod
    def read_package(cls, pkg_id):
        return (
            Session.query(PackageContributionActivity)
            .filter(PackageContributionActivity.package_id == pkg_id)
            .all()
        )
